from typing import Dict, Optional
import sass

class SCSSCompiler:
    def __init__(self, injected_vars: Optional[Dict[str, str]] = None):
        pass
        
    def compile(self, scss_content: str, namespace: str = "", param_overrides: Optional[Dict[str, str]] = None, component_name: str = "") -> str:
        if param_overrides:
            var_defs = []
            for key, value in param_overrides.items():
                var_name = key.replace("-", "_")
                if isinstance(value, str):
                    if not value.startswith(('rgb(', 'rgba(', '#', '"', "'")):
                        parts = value.split()
                        if len(parts) > 1:
                            is_css_multi_value = all(
                                part.rstrip('px%emremvwvhptcmmmininchpcexchvminvmax').replace('.', '').replace('-', '').isdigit() or 
                                part in ('auto', 'inherit', 'initial', 'unset')
                                for part in parts
                            )
                            if is_css_multi_value:
                                pass
                            else:
                                value = f'"{value}"'
                        else:
                            stripped = value.rstrip('px%emremvwvhptcmmmininchpcexchvminvmax')
                            try:
                                float(stripped)
                            except ValueError:
                                value = f'"{value}"'
                var_defs.append(f'${var_name}: {value};')
            scss_content = '\n'.join(var_defs) + '\n' + scss_content
        
        compiled_css = sass.compile(string=scss_content)
        
        if namespace:
            compiled_css = self._apply_namespace(compiled_css, namespace, component_name)
        
        return compiled_css
    
    def _apply_namespace(self, css: str, namespace: str, component_base_name: str = "") -> str:
        lines = []
        print(f"[NAMESPACE_INIT] component_base_name='{component_base_name}', namespace='{namespace}'")
        if not component_base_name:
            component_base_name = namespace.split('_')[-1] if '_' in namespace else namespace
            print(f"[NAMESPACE_INIT] Derived base name: '{component_base_name}'")
        
        for line in css.split('\n'):
            line = line.strip()
            if line and not line.startswith(('@', '/*', '}')):
                if '{' in line:
                    selector_part = line.split('{')[0].strip()
                    selectors = [s.strip() for s in selector_part.split(',')]
                    namespaced_selectors = []
                    for selector in selectors:
                        if selector:
                            selector_clean = selector.lstrip('.')
                            print(f"[NAMESPACE] selector='{selector}', base='{component_base_name}', namespace='{namespace}', match={selector_clean == component_base_name}")
                            if selector_clean == component_base_name:
                                namespaced_selectors.append(namespace)
                            elif selector_clean.startswith(component_base_name + '_'):
                                namespaced_selectors.append(selector_clean.replace(component_base_name, namespace, 1))
                            elif not selector_clean.startswith(namespace):
                                namespaced_selectors.append(f'{namespace}_{selector_clean}')
                            else:
                                namespaced_selectors.append(selector_clean)
                    line = ', '.join(namespaced_selectors) + ' {' + line.split('{', 1)[1]
            lines.append(line)
        return '\n'.join(lines)
    
    def compile_file(self, filepath: str, namespace: str = "", param_overrides: Optional[Dict[str, str]] = None, component_name: str = "") -> str:
        with open(filepath, 'r', encoding='utf-8') as f:
            scss_content = f.read()
        return self.compile(scss_content, namespace, param_overrides, component_name)



import bpy
import blf
import time

from .text_op import FontManager, font_manager
from .mouse_op import mouse_state

_text_input_instances = []
_draw_handle = None
_active_input_id = None
_keyboard_handler_running = False
_next_input_id = 0

class TextInputInstance:
    def __init__(self, container_id, placeholder="", font_name=None, size=20, pos=[50, 50], 
                 color=[1,1,1,1], mask=None, align_h='LEFT', align_v='TOP', 
                 cursor_color=[1,1,1,1], selection_color=[0.3,0.5,0.8,0.3]):
        global _next_input_id
        self.container_id = container_id
        self.id = _next_input_id
        _next_input_id += 1
        self.text = ""
        self.placeholder = placeholder
        self.font_name = font_name
        self.font_id = self._get_font_id()
        self.size = size
        self.position = pos
        self.color = color
        self.mask = mask
        self.align_h = align_h
        self.align_v = align_v
        self.cursor_color = cursor_color
        self.selection_color = selection_color
        
        self.cursor_pos = 0
        self.selection_start = None
        self.is_focused = False
        self.cursor_blink_time = 0.0
        self.show_cursor = True
        
        self.scroll_offset_x = 0
        self.scroll_offset_y = 0
        
        self._last_refresh = 0.0
        self._refresh_delay = 0.016
    
    def _get_font_id(self):
        if self.font_name and font_manager:
            try:
                font_id = font_manager.get_font_id(self.font_name)
                if font_id is not None and font_id >= 0:
                    return font_id
            except:
                pass
        return 0
    
    def refresh_font_id(self):
        self.font_id = self._get_font_id()
    
    def get_wrapped_lines(self):
        if not self.mask or self.mask[2] <= 0:
            return [self.text] if self.text else []
        
        container_width = self.mask[2]
        padding_total = 20
        available_width = max(container_width - padding_total, 50)
        
        blf.size(self.font_id, self.size)
        
        lines = []
        current_line = ""
        
        for char in self.text:
            if char == '\n':
                lines.append(current_line)
                current_line = ""
                continue
            
            test_line = current_line + char
            text_width, _ = blf.dimensions(self.font_id, test_line)
            
            if text_width > available_width:
                if current_line:
                    lines.append(current_line)
                    current_line = char
                else:
                    lines.append(char)
                    current_line = ""
            else:
                current_line = test_line
        
        if current_line:
            lines.append(current_line)
        
        return lines if lines else [""]
    
    def get_cursor_position_2d(self):
        lines = self.get_wrapped_lines()
        char_count = 0
        
        for line_idx, line in enumerate(lines):
            line_len = len(line)
            if char_count + line_len >= self.cursor_pos:
                col = self.cursor_pos - char_count
                return line_idx, col
            char_count += line_len
            if line_idx < len(lines) - 1:
                char_count += 1
        
        return len(lines) - 1, len(lines[-1]) if lines else 0
    
    def insert_text(self, text):
        if self.selection_start is not None:
            self.delete_selection()
        
        self.text = self.text[:self.cursor_pos] + text + self.text[self.cursor_pos:]
        self.cursor_pos += len(text)
        self._request_refresh()
    
    def delete_selection(self):
        if self.selection_start is None:
            return
        
        start = min(self.cursor_pos, self.selection_start)
        end = max(self.cursor_pos, self.selection_start)
        self.text = self.text[:start] + self.text[end:]
        self.cursor_pos = start
        self.selection_start = None
        self._request_refresh()
    
    def backspace(self):
        if self.selection_start is not None:
            self.delete_selection()
        elif self.cursor_pos > 0:
            self.text = self.text[:self.cursor_pos-1] + self.text[self.cursor_pos:]
            self.cursor_pos -= 1
            self._request_refresh()
    
    def delete(self):
        if self.selection_start is not None:
            self.delete_selection()
        elif self.cursor_pos < len(self.text):
            self.text = self.text[:self.cursor_pos] + self.text[self.cursor_pos+1:]
            self._request_refresh()
    
    def move_cursor_left(self, shift=False):
        if shift:
            if self.selection_start is None:
                self.selection_start = self.cursor_pos
        else:
            if self.selection_start is not None:
                self.cursor_pos = min(self.cursor_pos, self.selection_start)
                self.selection_start = None
                return
        
        if self.cursor_pos > 0:
            self.cursor_pos -= 1
        
        if not shift:
            self.selection_start = None
    
    def move_cursor_right(self, shift=False):
        if shift:
            if self.selection_start is None:
                self.selection_start = self.cursor_pos
        else:
            if self.selection_start is not None:
                self.cursor_pos = max(self.cursor_pos, self.selection_start)
                self.selection_start = None
                return
        
        if self.cursor_pos < len(self.text):
            self.cursor_pos += 1
        
        if not shift:
            self.selection_start = None
    
    def move_cursor_up(self, shift=False):
        if shift and self.selection_start is None:
            self.selection_start = self.cursor_pos
        
        lines = self.get_wrapped_lines()
        line_idx, col = self.get_cursor_position_2d()
        
        if line_idx > 0:
            prev_line = lines[line_idx - 1]
            new_col = min(col, len(prev_line))
            
            char_count = sum(len(lines[i]) + (1 if i < len(lines) - 1 else 0) for i in range(line_idx - 1))
            self.cursor_pos = char_count + new_col
        
        if not shift:
            self.selection_start = None
    
    def move_cursor_down(self, shift=False):
        if shift and self.selection_start is None:
            self.selection_start = self.cursor_pos
        
        lines = self.get_wrapped_lines()
        line_idx, col = self.get_cursor_position_2d()
        
        if line_idx < len(lines) - 1:
            next_line = lines[line_idx + 1]
            new_col = min(col, len(next_line))
            
            char_count = sum(len(lines[i]) + (1 if i < len(lines) - 1 else 0) for i in range(line_idx + 1))
            self.cursor_pos = char_count + new_col
        
        if not shift:
            self.selection_start = None
    
    def focus(self):
        global _active_input_id
        self.is_focused = True
        _active_input_id = self.id
        self.cursor_blink_time = time.time()
        self.show_cursor = True
    
    def blur(self):
        global _active_input_id
        self.is_focused = False
        if _active_input_id == self.id:
            _active_input_id = None
        self.selection_start = None
    
    def _request_refresh(self):
        self._last_refresh = time.time()
        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                area.tag_redraw()
    
    def should_refresh(self):
        return time.time() - self._last_refresh < self._refresh_delay

def draw_all_text_inputs():
    viewport_height = 0
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for region in area.regions:
                if region.type == 'WINDOW':
                    viewport_height = region.height
                    break
            break
    
    current_time = time.time()
    
    for instance in _text_input_instances:
        if instance.is_focused:
            if current_time - instance.cursor_blink_time > 0.5:
                instance.show_cursor = not instance.show_cursor
                instance.cursor_blink_time = current_time
        
        if instance.mask and instance.mask[2] > 0 and instance.mask[3] > 0:
            xmin = instance.mask[0]
            ymin = viewport_height - instance.mask[1] - instance.mask[3]
            xmax = instance.mask[0] + instance.mask[2]
            ymax = viewport_height - instance.mask[1]
            blf.clipping(instance.font_id, xmin, ymin, xmax, ymax)
            blf.enable(instance.font_id, blf.CLIPPING)
        
        blf.size(instance.font_id, instance.size)
        
        lines = instance.get_wrapped_lines()
        line_height = instance.size * 1.2
        
        display_text = instance.text if instance.text else instance.placeholder
        display_color = instance.color if instance.text else [c * 0.5 for c in instance.color]
        
        x_pos = instance.position[0]
        y_pos = instance.position[1]
        
        if instance.mask and instance.mask[2] > 0 and instance.mask[3] > 0:
            container_width = instance.mask[2]
            container_height = instance.mask[3]
            
            if instance.align_h == 'LEFT':
                x_pos = instance.mask[0] + 5
            elif instance.align_h == 'CENTER':
                x_pos = instance.mask[0] + container_width / 2
            elif instance.align_h == 'RIGHT':
                x_pos = instance.mask[0] + container_width - 5
            
            if instance.align_v == 'TOP':
                y_pos = instance.mask[1] + 5
            elif instance.align_v == 'CENTER':
                total_text_height = len(lines) * line_height
                y_pos = instance.mask[1] + (container_height - total_text_height) / 2
            elif instance.align_v == 'BOTTOM':
                total_text_height = len(lines) * line_height
                y_pos = instance.mask[1] + container_height - total_text_height - 5
        
        for line_idx, line in enumerate(lines):
            line_y = y_pos + line_idx * line_height
            text_baseline_offset = (line_height - instance.size) / 2
            flipped_y = viewport_height - line_y - instance.size - text_baseline_offset
            selection_y = viewport_height - line_y - line_height
            
            if instance.selection_start is not None and instance.text:
                start = min(instance.cursor_pos, instance.selection_start)
                end = max(instance.cursor_pos, instance.selection_start)
                
                char_count = sum(len(lines[i]) + (1 if i < len(lines) - 1 else 0) for i in range(line_idx))
                line_start = char_count
                line_end = char_count + len(line)
                
                if start < line_end and end > line_start:
                    sel_start_in_line = max(0, start - line_start)
                    sel_end_in_line = min(len(line), end - line_start)
                    
                    before_sel = line[:sel_start_in_line]
                    selected = line[sel_start_in_line:sel_end_in_line]
                    
                    before_width, _ = blf.dimensions(instance.font_id, before_sel)
                    sel_width, _ = blf.dimensions(instance.font_id, selected)
                    
                    import gpu
                    from gpu_extras.batch import batch_for_shader
                    
                    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
                    
                    sel_x = x_pos + before_width
                    sel_height = line_height
                    
                    vertices = (
                        (sel_x, selection_y),
                        (sel_x + sel_width, selection_y),
                        (sel_x, selection_y + sel_height),
                        (sel_x + sel_width, selection_y + sel_height),
                    )
                    
                    indices = ((0, 1, 2), (1, 2, 3))
                    
                    batch = batch_for_shader(shader, 'TRIS', {"pos": vertices}, indices=indices)
                    shader.bind()
                    shader.uniform_float("color", instance.selection_color)
                    batch.draw(shader)
            
            blf.position(instance.font_id, x_pos, flipped_y, 0)
            blf.color(instance.font_id, *display_color)
            blf.draw(instance.font_id, line)
        
        if instance.is_focused and instance.show_cursor:
            line_idx, col = instance.get_cursor_position_2d()
            
            if line_idx < len(lines):
                line = lines[line_idx]
                text_before_cursor = line[:col]
                cursor_x_offset, _ = blf.dimensions(instance.font_id, text_before_cursor)
                
                cursor_x = x_pos + cursor_x_offset
                cursor_y_line = y_pos + line_idx * line_height
                cursor_y_flipped = viewport_height - cursor_y_line - line_height
                
                import gpu
                from gpu_extras.batch import batch_for_shader
                
                shader = gpu.shader.from_builtin('UNIFORM_COLOR')
                
                cursor_height = line_height
                cursor_width = 2
                vertices = (
                    (cursor_x, cursor_y_flipped),
                    (cursor_x + cursor_width, cursor_y_flipped),
                    (cursor_x, cursor_y_flipped + cursor_height),
                    (cursor_x + cursor_width, cursor_y_flipped + cursor_height),
                )
                
                indices = ((0, 1, 2), (1, 2, 3))
                
                batch = batch_for_shader(shader, 'TRIS', {"pos": vertices}, indices=indices)
                shader.bind()
                shader.uniform_float("color", instance.cursor_color)
                batch.draw(shader)
        
        if instance.mask and instance.mask[2] > 0 and instance.mask[3] > 0:
            blf.disable(instance.font_id, blf.CLIPPING)

class KeyboardHandler(bpy.types.Operator):
    bl_idname = "xwz.text_input_keyboard"
    bl_label = "Text Input Keyboard Handler"
    
    def invoke(self, context, event):
        global _keyboard_handler_running
        _keyboard_handler_running = True
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}
    
    def modal(self, context, event):
        global _active_input_id, _text_input_instances, _keyboard_handler_running
        
        if not _text_input_instances:
            _keyboard_handler_running = False
            return {'CANCELLED'}
        
        if _active_input_id is None:
            return {'PASS_THROUGH'}
        
        active_input = None
        for instance in _text_input_instances:
            if instance.id == _active_input_id:
                active_input = instance
                break
        
        if not active_input:
            return {'PASS_THROUGH'}
        
        if event.type in {'LEFTMOUSE', 'RIGHTMOUSE'} and event.value == 'PRESS':
            return {'PASS_THROUGH'}
        
        if event.type == 'ESC' and event.value == 'PRESS':
            active_input.blur()
            return {'RUNNING_MODAL'}
        
        if event.type == 'RET' and event.value == 'PRESS':
            if event.shift:
                active_input.insert_text('\n')
            else:
                active_input.blur()
            return {'RUNNING_MODAL'}
        
        if event.type == 'BACK_SPACE' and event.value == 'PRESS':
            active_input.backspace()
            return {'RUNNING_MODAL'}
        
        if event.type == 'DEL' and event.value == 'PRESS':
            active_input.delete()
            return {'RUNNING_MODAL'}
        
        if event.type == 'LEFT_ARROW' and event.value == 'PRESS':
            active_input.move_cursor_left(event.shift)
            return {'RUNNING_MODAL'}
        
        if event.type == 'RIGHT_ARROW' and event.value == 'PRESS':
            active_input.move_cursor_right(event.shift)
            return {'RUNNING_MODAL'}
        
        if event.type == 'UP_ARROW' and event.value == 'PRESS':
            active_input.move_cursor_up(event.shift)
            return {'RUNNING_MODAL'}
        
        if event.type == 'DOWN_ARROW' and event.value == 'PRESS':
            active_input.move_cursor_down(event.shift)
            return {'RUNNING_MODAL'}
        
        if event.type == 'HOME' and event.value == 'PRESS':
            if event.ctrl:
                active_input.cursor_pos = 0
            else:
                lines = active_input.get_wrapped_lines()
                line_idx, col = active_input.get_cursor_position_2d()
                char_count = sum(len(lines[i]) + (1 if i < len(lines) - 1 else 0) for i in range(line_idx))
                active_input.cursor_pos = char_count
            
            if not event.shift:
                active_input.selection_start = None
            return {'RUNNING_MODAL'}
        
        if event.type == 'END' and event.value == 'PRESS':
            if event.ctrl:
                active_input.cursor_pos = len(active_input.text)
            else:
                lines = active_input.get_wrapped_lines()
                line_idx, col = active_input.get_cursor_position_2d()
                char_count = sum(len(lines[i]) + (1 if i < len(lines) - 1 else 0) for i in range(line_idx))
                active_input.cursor_pos = char_count + len(lines[line_idx])
            
            if not event.shift:
                active_input.selection_start = None
            return {'RUNNING_MODAL'}
        
        if event.type == 'A' and event.value == 'PRESS' and event.ctrl:
            active_input.selection_start = 0
            active_input.cursor_pos = len(active_input.text)
            return {'RUNNING_MODAL'}
        
        if event.type == 'C' and event.value == 'PRESS' and event.ctrl:
            if active_input.selection_start is not None:
                start = min(active_input.cursor_pos, active_input.selection_start)
                end = max(active_input.cursor_pos, active_input.selection_start)
                selected_text = active_input.text[start:end]
                context.window_manager.clipboard = selected_text
            return {'RUNNING_MODAL'}
        
        if event.type == 'X' and event.value == 'PRESS' and event.ctrl:
            if active_input.selection_start is not None:
                start = min(active_input.cursor_pos, active_input.selection_start)
                end = max(active_input.cursor_pos, active_input.selection_start)
                selected_text = active_input.text[start:end]
                context.window_manager.clipboard = selected_text
                active_input.delete_selection()
            return {'RUNNING_MODAL'}
        
        if event.type == 'V' and event.value == 'PRESS' and event.ctrl:
            clipboard_text = context.window_manager.clipboard
            if clipboard_text:
                active_input.insert_text(clipboard_text)
            return {'RUNNING_MODAL'}
        
        if event.value == 'PRESS' and event.ascii:
            if not event.ctrl and not event.alt:
                active_input.insert_text(event.ascii)
                return {'RUNNING_MODAL'}
        
        return {'PASS_THROUGH'}

class CreateTextInputOP(bpy.types.Operator):
    bl_idname = "xwz.create_text_input"
    bl_label = "Create Text Input"
    
    container_id: bpy.props.StringProperty(name="Container ID", default="root")
    placeholder: bpy.props.StringProperty(name="Placeholder", default="Type here...")
    font_name: bpy.props.StringProperty(name="Font Name", default="")
    size: bpy.props.IntProperty(name="Size", default=20, min=1, max=200)
    x_pos: bpy.props.IntProperty(name="X Position", default=50)
    y_pos: bpy.props.IntProperty(name="Y Position", default=50)
    color: bpy.props.FloatVectorProperty(name="Color", subtype='COLOR', size=4, default=(1.0, 1.0, 1.0, 1.0))
    cursor_color: bpy.props.FloatVectorProperty(name="Cursor Color", subtype='COLOR', size=4, default=(1.0, 1.0, 1.0, 1.0))
    selection_color: bpy.props.FloatVectorProperty(name="Selection Color", subtype='COLOR', size=4, default=(0.3, 0.5, 0.8, 0.3))
    mask_x: bpy.props.IntProperty(name="Mask X", default=0)
    mask_y: bpy.props.IntProperty(name="Mask Y", default=0)
    mask_width: bpy.props.IntProperty(name="Mask Width", default=0)
    mask_height: bpy.props.IntProperty(name="Mask Height", default=0)
    align_h: bpy.props.EnumProperty(
        name="Horizontal Align",
        items=[('LEFT', 'Left', ''), ('CENTER', 'Center', ''), ('RIGHT', 'Right', '')],
        default='LEFT'
    )
    align_v: bpy.props.EnumProperty(
        name="Vertical Align",
        items=[('TOP', 'Top', ''), ('CENTER', 'Center', ''), ('BOTTOM', 'Bottom', '')],
        default='TOP'
    )
    
    def execute(self, context):
        global _draw_handle, _text_input_instances, _keyboard_handler_running
        
        mask = None
        if self.mask_width > 0 and self.mask_height > 0:
            mask = [self.mask_x, self.mask_y, self.mask_width, self.mask_height]
        
        font_to_use = self.font_name if self.font_name else None
        if font_to_use and font_to_use not in font_manager.get_available_fonts():
            print(f"Warning: Font '{font_to_use}' not found, using default")
            font_to_use = None
        
        new_instance = TextInputInstance(
            container_id=self.container_id,
            placeholder=self.placeholder,
            font_name=font_to_use,
            size=self.size,
            pos=[self.x_pos, self.y_pos],
            color=list(self.color),
            mask=mask,
            align_h=self.align_h,
            align_v=self.align_v,
            cursor_color=list(self.cursor_color),
            selection_color=list(self.selection_color)
        )
        
        new_instance.refresh_font_id()
        
        _text_input_instances.append(new_instance)
        
        if _draw_handle is None:
            _draw_handle = bpy.types.SpaceView3D.draw_handler_add(
                draw_all_text_inputs, (), 'WINDOW', 'POST_PIXEL')
        
        if not _keyboard_handler_running:
            bpy.ops.xwz.text_input_keyboard('INVOKE_DEFAULT')
        
        context.area.tag_redraw()
        self.report({'INFO'}, f"Created text input instance #{new_instance.id}")
        return {'FINISHED'}

class RemoveTextInputOP(bpy.types.Operator):
    bl_idname = "xwz.remove_text_input"
    bl_label = "Remove Text Input"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    
    def execute(self, context):
        global _draw_handle, _text_input_instances, _active_input_id
        
        for i, instance in enumerate(_text_input_instances):
            if instance.id == self.instance_id:
                if _active_input_id == instance.id:
                    _active_input_id = None
                _text_input_instances.pop(i)
                self.report({'INFO'}, f"Removed text input instance #{self.instance_id}")
                break
        else:
            self.report({'ERROR'}, f"Text input instance #{self.instance_id} not found")
            return {'CANCELLED'}
        
        if not _text_input_instances and _draw_handle is not None:
            bpy.types.SpaceView3D.draw_handler_remove(_draw_handle, 'WINDOW')
            _draw_handle = None
        
        context.area.tag_redraw()
        return {'FINISHED'}

class ClearTextInputsOP(bpy.types.Operator):
    bl_idname = "xwz.clear_text_inputs"
    bl_label = "Clear All Text Inputs"
    
    def execute(self, context):
        global _draw_handle, _text_input_instances, _active_input_id, _keyboard_handler_running, _next_input_id
        
        _text_input_instances.clear()
        _active_input_id = None
        _keyboard_handler_running = False
        _next_input_id = 0
        
        if _draw_handle is not None:
            bpy.types.SpaceView3D.draw_handler_remove(_draw_handle, 'WINDOW')
            _draw_handle = None
        
        context.area.tag_redraw()
        return {'FINISHED'}

class FocusTextInputOP(bpy.types.Operator):
    bl_idname = "xwz.focus_text_input"
    bl_label = "Focus Text Input"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    
    def execute(self, context):
        global _text_input_instances
        
        for instance in _text_input_instances:
            if instance.id == self.instance_id:
                for other in _text_input_instances:
                    if other.id != self.instance_id:
                        other.blur()
                
                instance.focus()
                context.area.tag_redraw()
                return {'FINISHED'}
        
        self.report({'ERROR'}, f"Text input instance #{self.instance_id} not found")
        return {'CANCELLED'}

class BlurTextInputOP(bpy.types.Operator):
    bl_idname = "xwz.blur_text_input"
    bl_label = "Blur Text Input"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    
    def execute(self, context):
        global _text_input_instances
        
        for instance in _text_input_instances:
            if instance.id == self.instance_id:
                instance.blur()
                context.area.tag_redraw()
                return {'FINISHED'}
        
        self.report({'ERROR'}, f"Text input instance #{self.instance_id} not found")
        return {'CANCELLED'}

class GetTextInputValueOP(bpy.types.Operator):
    bl_idname = "xwz.get_text_input_value"
    bl_label = "Get Text Input Value"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    
    def execute(self, context):
        global _text_input_instances
        
        for instance in _text_input_instances:
            if instance.id == self.instance_id:
                print(f"Text input #{self.instance_id} value: {instance.text}")
                return {'FINISHED'}
        
        self.report({'ERROR'}, f"Text input instance #{self.instance_id} not found")
        return {'CANCELLED'}

class SetTextInputValueOP(bpy.types.Operator):
    bl_idname = "xwz.set_text_input_value"
    bl_label = "Set Text Input Value"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    text: bpy.props.StringProperty(name="Text", default="")
    
    def execute(self, context):
        global _text_input_instances
        
        for instance in _text_input_instances:
            if instance.id == self.instance_id:
                instance.text = self.text
                instance.cursor_pos = len(self.text)
                instance.selection_start = None
                instance._request_refresh()
                return {'FINISHED'}
        
        self.report({'ERROR'}, f"Text input instance #{self.instance_id} not found")
        return {'CANCELLED'}

class UpdateTextInputOP(bpy.types.Operator):
    bl_idname = "xwz.update_text_input"
    bl_label = "Update Text Input"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    placeholder: bpy.props.StringProperty(name="Placeholder", default="")
    font_name: bpy.props.StringProperty(name="Font Name", default="")
    size: bpy.props.IntProperty(name="Size", default=-1, min=-1, max=200)
    x_pos: bpy.props.IntProperty(name="X Position", default=-999999)
    y_pos: bpy.props.IntProperty(name="Y Position", default=-999999)
    color: bpy.props.FloatVectorProperty(name="Color", subtype='COLOR', size=4, default=(-1, -1, -1, -1))
    cursor_color: bpy.props.FloatVectorProperty(name="Cursor Color", subtype='COLOR', size=4, default=(-1, -1, -1, -1))
    selection_color: bpy.props.FloatVectorProperty(name="Selection Color", subtype='COLOR', size=4, default=(-1, -1, -1, -1))
    mask_x: bpy.props.IntProperty(name="Mask X", default=-999999)
    mask_y: bpy.props.IntProperty(name="Mask Y", default=-999999)
    mask_width: bpy.props.IntProperty(name="Mask Width", default=-1)
    mask_height: bpy.props.IntProperty(name="Mask Height", default=-1)
    align_h: bpy.props.EnumProperty(
        name="Horizontal Align",
        items=[('__NOCHANGE__', 'No Change', ''), ('LEFT', 'Left', ''), ('CENTER', 'Center', ''), ('RIGHT', 'Right', '')],
        default='__NOCHANGE__'
    )
    align_v: bpy.props.EnumProperty(
        name="Vertical Align",
        items=[('__NOCHANGE__', 'No Change', ''), ('TOP', 'Top', ''), ('CENTER', 'Center', ''), ('BOTTOM', 'Bottom', '')],
        default='__NOCHANGE__'
    )
    
    def execute(self, context):
        for instance in _text_input_instances:
            if instance.id == self.instance_id:
                updated_props = []
                
                if self.placeholder:
                    instance.placeholder = self.placeholder
                    updated_props.append('placeholder')
                
                if self.font_name and self.font_name in font_manager.get_available_fonts():
                    instance.font_name = self.font_name
                    instance.refresh_font_id()
                    updated_props.append('font_name')
                
                if self.size != -1:
                    instance.size = self.size
                    updated_props.append('size')
                
                if self.x_pos != -999999 or self.y_pos != -999999:
                    if self.x_pos != -999999:
                        instance.position[0] = self.x_pos
                    if self.y_pos != -999999:
                        instance.position[1] = self.y_pos
                    updated_props.append('position')
                
                if any(c != -1 for c in self.color):
                    for i in range(4):
                        if self.color[i] != -1:
                            instance.color[i] = self.color[i]
                    updated_props.append('color')
                
                if any(c != -1 for c in self.cursor_color):
                    for i in range(4):
                        if self.cursor_color[i] != -1:
                            instance.cursor_color[i] = self.cursor_color[i]
                    updated_props.append('cursor_color')
                
                if any(c != -1 for c in self.selection_color):
                    for i in range(4):
                        if self.selection_color[i] != -1:
                            instance.selection_color[i] = self.selection_color[i]
                    updated_props.append('selection_color')
                
                if (self.mask_x != -999999 or self.mask_y != -999999 or 
                    self.mask_width != -1 or self.mask_height != -1):
                    current_mask = instance.mask or [0, 0, 0, 0]
                    new_mask = [
                        self.mask_x if self.mask_x != -999999 else current_mask[0],
                        self.mask_y if self.mask_y != -999999 else current_mask[1],
                        self.mask_width if self.mask_width != -1 else current_mask[2],
                        self.mask_height if self.mask_height != -1 else current_mask[3]
                    ]
                    instance.mask = new_mask if new_mask[2] > 0 and new_mask[3] > 0 else None
                    updated_props.append('mask')
                
                if self.align_h != '__NOCHANGE__':
                    instance.align_h = self.align_h
                    updated_props.append('align_h')
                
                if self.align_v != '__NOCHANGE__':
                    instance.align_v = self.align_v
                    updated_props.append('align_v')
                
                if updated_props:
                    instance._request_refresh()
                    self.report({'INFO'}, f"Updated text input #{self.instance_id}: {', '.join(updated_props)}")
                else:
                    self.report({'INFO'}, f"No properties specified to update for text input #{self.instance_id}")
                
                return {'FINISHED'}
        
        self.report({'ERROR'}, f"Text input instance #{self.instance_id} not found")
        return {'CANCELLED'}

def register():
    bpy.utils.register_class(KeyboardHandler)
    bpy.utils.register_class(CreateTextInputOP)
    bpy.utils.register_class(RemoveTextInputOP)
    bpy.utils.register_class(ClearTextInputsOP)
    bpy.utils.register_class(FocusTextInputOP)
    bpy.utils.register_class(BlurTextInputOP)
    bpy.utils.register_class(GetTextInputValueOP)
    bpy.utils.register_class(SetTextInputValueOP)
    bpy.utils.register_class(UpdateTextInputOP)

def unregister():
    global _draw_handle, _text_input_instances, _active_input_id, _keyboard_handler_running, _next_input_id
    
    _text_input_instances.clear()
    _active_input_id = None
    _keyboard_handler_running = False
    _next_input_id = 0
    
    if _draw_handle is not None:
        bpy.types.SpaceView3D.draw_handler_remove(_draw_handle, 'WINDOW')
        _draw_handle = None
    
    bpy.utils.unregister_class(KeyboardHandler)
    bpy.utils.unregister_class(CreateTextInputOP)
    bpy.utils.unregister_class(RemoveTextInputOP)
    bpy.utils.unregister_class(ClearTextInputsOP)
    bpy.utils.unregister_class(FocusTextInputOP)
    bpy.utils.unregister_class(BlurTextInputOP)
    bpy.utils.unregister_class(GetTextInputValueOP)
    bpy.utils.unregister_class(SetTextInputValueOP)
    bpy.utils.unregister_class(UpdateTextInputOP)



class TextInputExtractor:
    def __init__(self, ui, json_data):
        self.ui = ui
        self.json_data = json_data
        self.text_input_blocks = {}
        self.flat_index = 0
        self._extract_text_inputs(self.ui.theme.root)
    
    def _extract_text_inputs(self, container):
        if container.data != '' and container.data.startswith('<INPUT>'):
            placeholder = ""
            if '|' in container.data:
                parts = container.data.split('|', 1)
                if len(parts) > 1:
                    placeholder = parts[1].strip()
            
            self.text_input_blocks[container.id] = {
                'container_id': container.id,
                'placeholder': placeholder,
                'font': container.font if container.font != '' else self.ui.theme.default_font,
                'x_pos': int(self.json_data[self.flat_index]['position'][0] + container.style.text_x),
                'y_pos': int(self.json_data[self.flat_index]['position'][1] + container.style.text_y),
                'text_scale': int(container.style.text_scale),
                'text_color': container.style.text_color,
                'text_color_1': container.style.text_color_1,
                'text_color_gradient_rot': container.style.text_color_gradient_rot,
                'mask_x': int(self.json_data[self.flat_index]['position'][0]),
                'mask_y': int(self.json_data[self.flat_index]['position'][1]),
                'mask_width': int(self.json_data[self.flat_index]['size'][0]),
                'mask_height': int(self.json_data[self.flat_index]['size'][1]),
                'align_h': container.style.text_align_h,
                'align_v': container.style.text_align_v
            }
        
        self.flat_index += 1
        for child in container.children:
            self._extract_text_inputs(child)



from __future__ import annotations
from typing import Optional, List

class Container(): 
    def __init__(self): 
        self.id       : str                       = ""
        self.parent   : Optional[Container]       = []
        self.children : Optional[List[Container]] = []

        self.style : Optional[str] = ""
        self.data  : Optional[str] = ""
        self.img   : Optional[str] = ""
        self.text  : Optional[str] = ""
        self.font  : Optional[str] = ""

        self.layer   : int   = 0
        self.passive : bool  = False

        self.click         : List  = []
        self.toggle        : List  = []
        self.scroll        : List  = []
        self.hover         : List  = []
        self.hoverout      : List  = []
        
        self._toggle_value : bool  = False
        self._toggled      : bool  = False
        self._clicked      : bool  = False
        self._hovered      : bool  = False
        self._prev_toggled : bool  = False
        self._prev_clicked : bool  = False
        self._prev_hovered : bool  = False
        self._scroll_value : float = 0.0
        
        self._dirty        : bool  = False
    
    def __getattr__(self, name):
        """
        Enable property-based access to:
        1. Child containers by their id (e.g., app.theme.root.bg.body)
        2. Style properties (e.g., container.text_scale delegates to container.style.text_scale)
        """
        # Avoid infinite recursion by checking if we're accessing core attributes
        if name in ('children', 'style', '__dict__'):
            raise AttributeError(f"'Container' object has no attribute '{name}'")
        
        # First, search for a child container with matching id
        try:
            children = object.__getattribute__(self, 'children')
            for child in children:
                if child.id == name or child.id.endswith(f"_{name}"):
                    return child
        except AttributeError:
            pass
        
        # If not found in children, try to get from style object
        try:
            style = object.__getattribute__(self, 'style')
            if style and hasattr(style, name):
                return getattr(style, name)
        except AttributeError:
            pass
        
        # If not found anywhere, raise AttributeError
        raise AttributeError(f"'Container' object has no attribute or child named '{name}'")
    
    def __setattr__(self, name, value):
        """
        Enable property-based setting of style properties.
        If the attribute doesn't exist on Container but exists on style,
        delegate to the style object.
        """
        # List of Container's own attributes that should be set directly
        container_attrs = {
            'id', 'parent', 'children', 'style', 'data', 'img', 'text', 'font',
            'layer', 'passive', 'click', 'toggle', 'scroll', 'hover', 'hoverout',
            '_toggle_value', '_toggled', '_clicked', '_hovered',
            '_prev_toggled', '_prev_clicked', '_prev_hovered', '_scroll_value', '_dirty'
        }
        
        # If it's a Container attribute, set it directly
        if name in container_attrs:
            object.__setattr__(self, name, value)
        # Otherwise, try to set it on the style object if it exists there
        else:
            try:
                style = object.__getattribute__(self, 'style')
                if style and hasattr(style, name):
                    setattr(style, name, value)
                else:
                    # If style doesn't have it either, set it on container
                    object.__setattr__(self, name, value)
            except AttributeError:
                # If style doesn't exist yet, set it on container
                object.__setattr__(self, name, value)
    
    def mark_dirty(self):
        self._dirty = True
    
    def get_by_id(self, target_id):
        if self.id == target_id or self.id.endswith(f"_{target_id}"):
            return self
        if self.children:
            for child in self.children:
                result = child.get_by_id(target_id)
                if result:
                    return result
        return None

class ContainerDefault():
    def __init__(self): 
        self.id    = None
        self.style = None

        self.parent   = None
        self.children = []

        self.click         = []
        self.toggle        = []
        self.scroll        = []
        self.hover         = []
        self.hoverout      = []
        self._toggle_value = False
        self._toggled      = False
        self._clicked      = False
        self._hovered      = False
        self._prev_toggled = False
        self._prev_clicked = False
        self._prev_hovered = False
        self._scroll_value = 0.0

        self.display      = True
        self.overflow     = False
        self.data         = ""
        self.img          = ""
        self.aspect_ratio = False
        self.text         = ""
        self.font         = 'NeueMontreal-Regular'

        self.layer   = 0
        self.passive = False

        self.x = 0.0
        self.y = 0.0

        self.width  = 100.0
        self.height = 100.0

        self.color              = [0.0, 0.0, 0.0, 1.0]
        self.color_1            = [0.0, 0.0, 0.0, 0.0]
        self.color_gradient_rot = 0.0
        
        self.hover_color              = [0.0, 0.0, 0.0, -1.0]
        self.hover_color_1            = [0.0, 0.0, 0.0, 0.0]
        self.hover_color_gradient_rot = 0.0

        self.click_color              = [0.0, 0.0, 0.0, -1.0]
        self.click_color_1            = [0.0, 0.0, 0.0, 0.0]
        self.click_color_gradient_rot = 0.0

        self.toggle_color              = [0.0, 0.0, 0.0, -1.0]
        self.toggle_color_1            = [0.0, 0.0, 0.0, 0.0]
        self.toggle_color_gradient_rot = 0.0

        self.border_color              = [0.0, 0.0, 0.0, 0.0]
        self.border_color_1            = [0.0, 0.0, 0.0, 0.0]
        self.border_color_gradient_rot = 0.0
        self.border_radius             = 0.0
        self.border_width              = 0.0
        
        self.text_color              = [1.0, 1.0, 1.0, 1.0]
        self.text_color_1            = [0.0, 0.0, 0.0, 0.0]
        self.text_color_gradient_rot = 0.0
        self.text_scale              = 12.0
        self.text_x                  = 0.0
        self.text_y                  = 0.0
        
        self.box_shadow_color  = [0.0, 0.0, 0.0, 0.0]
        self.box_shadow_offset = [0.0, 0.0, 0.0]
        self.box_shadow_blur   = 0.0
        
container_default = ContainerDefault()


class ImageExtractor():
    def __init__(self, ui, json_data):
        self.ui                    = ui
        self.json_data             = json_data
        self.image_blocks          = {}
        self.image_blocks_relative = {}
        self.flat_index            = 0
        self._extract_images(self.ui.theme.root)
    def _extract_images(self, container):
        if container.img != '':
            self.image_blocks[container.id] = {
                'container_id': container.id,
                'image_name'  : container.img,
                'x_pos'       : int(self.json_data[self.flat_index]['position'][0]),
                'y_pos'       : int(self.json_data[self.flat_index]['position'][1]),
                'width'       : int(self.json_data[self.flat_index]['size'][0]),
                'height'      : int(self.json_data[self.flat_index]['size'][1]),
                "mask_x"      : int(self.json_data[self.flat_index]['position'][0]),
                "mask_y"      : int(self.json_data[self.flat_index]['position'][1]),
                "mask_width"  : int(self.json_data[self.flat_index]['size'][0]),
                "mask_height" : int(self.json_data[self.flat_index]['size'][1]),
                "aspect_ratio": container.style.aspect_ratio,
                'align_h'     : container.style.img_align_h,
                'align_v'     : container.style.img_align_v
            }
        self.flat_index += 1
        for child in container.children:
            self._extract_images(child)




import bpy
import gpu
import os
import time
import moderngl as mgl
from .components.container import container_default
import numpy as np

from gpu_extras.batch import batch_for_shader
from bpy.types import Operator, Panel

from .scroll_op import scroll_state, XWZ_OT_scroll, XWZ_OT_scroll_launch
from .mouse_op import mouse_state, XWZ_OT_mouse, XWZ_OT_mouse_launch
from .parser_op import XWZ_OT_ui_parser
from . import parser_op

_render_data = None
_modal_timer = None

class RenderPipeline:
    def __init__(self):
        self.mgl_context     = None
        self.compute_shader  = None
        self.mouse_buffer    = None
        self.container_buffer = None
        self.viewport_buffer = None
        self.output_texture  = None
        self.blender_texture = None
        self.gpu_shader      = None
        self.batch           = None
        self.draw_handler    = None
        self.running         = False
        self.mouse_pos       = [0.5, 0.5]
        self.start_time      = time.time()
        self.texture_size    = (1920, 1080)
        self.click_value     = 0.0
        self.scroll_callback_registered = False
        self.mouse_callback_registered = False
        self.region_size     = (1, 1)
        self.container_data  = []
        self.frame_times     = []
        self.compute_fps     = 0.0
        self.last_frame_time = time.perf_counter()
        self.needs_texture_update = True
        self.texture_needs_readback = True  # Flag for draw_texture to know if readback is needed
        self.last_mouse_pos = [0.5, 0.5]
        self.last_click_value = 0.0
        self.last_scroll_value = 0.0
        self.last_container_update = 0
        self.conf_path = 'xwz.ui.toml'
    def _safe_release_moderngl_object(self, obj):
        """Safely release a ModernGL object, checking if it's valid first"""
        if obj and hasattr(obj, 'mglo'):
            try:
                if type(obj.mglo).__name__ != 'InvalidObject':
                    obj.release()
                return True
            except Exception:
                return False
        return False
    def load_shader_file(self, filename):
        # Shader files are in the puree package directory
        package_dir = os.path.dirname(os.path.abspath(__file__))
        shader_path = os.path.join(package_dir, "shaders", filename)
        try:
            with open(shader_path, 'r') as f:
                return f.read()
        except Exception:
            return None
    def load_container_data(self):
        try:  
            wm = bpy.context.window_manager
            bpy.ops.xwz.parse_app_ui(conf_path=wm.xwz_ui_conf_path)
            self.container_data = parser_op._container_json_data
            return True
        except Exception:
            return False
    def init_moderngl_context(self):
        try:
            self.mgl_context = mgl.get_context()
            self.mgl_context.gc_mode = 'context_gc'
            return True
        except Exception:
            return False
    def create_compute_shader(self):
        shader_source = self.load_shader_file("container.glsl")
        if not shader_source:
            return False
        try:
            self.compute_shader = self.mgl_context.compute_shader(shader_source)
            return True
        except Exception:
            return False
    def create_buffers_and_textures(self):
        try:
            mouse_data = np.array([0.5, 0.5, 0.0, 0.0, 0.0, 0.0], dtype=np.float32)
            self.mouse_buffer = self.mgl_context.buffer(mouse_data.tobytes())
            
            container_array = []
            for i, container in enumerate(self.container_data):
                container_struct = [
                    int(container.get('display', False)),
                    container.get('position', [0, 0])[0], container.get('position', [0, 0])[1],
                    container.get('size', [100, 100])[0], container.get('size', [100, 100])[1],
                    container.get('color', [1, 1, 1, 1])[0], container.get('color', [1, 1, 1, 1])[1], 
                    container.get('color', [1, 1, 1, 1])[2], container.get('color', [1, 1, 1, 1])[3],
                    container.get('color_1', [1, 1, 1, 1])[0], container.get('color_1', [1, 1, 1, 1])[1], 
                    container.get('color_1', [1, 1, 1, 1])[2], container.get('color_1', [1, 1, 1, 1])[3],
                    container.get('color_gradient_rot', 0.0),
                    container.get('hover_color', container_default.hover_color)[0], container.get('hover_color', container_default.hover_color)[1], 
                    container.get('hover_color', container_default.hover_color)[2], container.get('hover_color', container_default.hover_color)[3],
                    container.get('hover_color_1', container_default.hover_color_1)[0], container.get('hover_color_1', container_default.hover_color_1)[1], 
                    container.get('hover_color_1', container_default.hover_color_1)[2], container.get('hover_color_1', container_default.hover_color_1)[3],
                    container.get('hover_color_gradient_rot', 0.0),
                    container.get('click_color', container_default.click_color)[0], container.get('click_color', container_default.click_color)[1], 
                    container.get('click_color', container_default.click_color)[2], container.get('click_color', container_default.click_color)[3],
                    container.get('click_color_1', container_default.click_color_1)[0], container.get('click_color_1', container_default.click_color_1)[1], 
                    container.get('click_color_1', container_default.click_color_1)[2], container.get('click_color_1', container_default.click_color_1)[3],
                    container.get('click_color_gradient_rot', 0.0),
                    container.get('border_color', [1, 1, 1, 1])[0], container.get('border_color', [1, 1, 1, 1])[1], 
                    container.get('border_color', [1, 1, 1, 1])[2], container.get('border_color', [1, 1, 1, 1])[3],
                    container.get('border_color_1', [1, 1, 1, 1])[0], container.get('border_color_1', [1, 1, 1, 1])[1], 
                    container.get('border_color_1', [1, 1, 1, 1])[2], container.get('border_color_1', [1, 1, 1, 1])[3],
                    container.get('border_color_gradient_rot', 0.0),
                    container.get('border_radius', 0.0),
                    container.get('border_width', 0.0),
                    container.get('parent', -1),
                    int(container.get('overflow', False)),
                    container.get('box_shadow_offset', [0, 0, 0])[0], container.get('box_shadow_offset', [0, 0, 0])[1], 
                    container.get('box_shadow_offset', [0, 0, 0])[2],
                    container.get('box_shadow_blur', 0.0),
                    container.get('box_shadow_color', [0, 0, 0, 0])[0], container.get('box_shadow_color', [0, 0, 0, 0])[1], 
                    container.get('box_shadow_color', [0, 0, 0, 0])[2], container.get('box_shadow_color', [0, 0, 0, 0])[3],
                    int(container.get('passive', False))
                ]
                container_array.extend(container_struct)
            
            container_data_np = np.array(container_array, dtype=np.float32)
            self.container_buffer = self.mgl_context.buffer(container_data_np.tobytes())
            
            viewport_data = np.array([self.region_size[0], self.region_size[1], len(self.container_data)], dtype=np.float32)
            self.viewport_buffer = self.mgl_context.buffer(viewport_data.tobytes())
            
            self.texture_size = self.region_size
            
            self.output_texture = self.mgl_context.texture(
                self.texture_size, 
                4
            )
            self.output_texture.filter = (mgl.NEAREST, mgl.NEAREST)
            return True
        except Exception:
            return False
    def create_blender_gpu_shader(self):
        vert_source = self.load_shader_file("vertex.glsl")
        frag_source = self.load_shader_file("fragment.glsl")
        
        if not (vert_source and frag_source):
            return False
            
        try:
            shader_info = gpu.types.GPUShaderCreateInfo()
            
            shader_info.vertex_in(0, 'VEC2', 'position')
            shader_info.vertex_in(1, 'VEC2', 'texCoord_0')
            
            interface = gpu.types.GPUStageInterfaceInfo("default_interface")
            interface.smooth('VEC2', 'fragTexCoord')
            shader_info.vertex_out(interface)
            
            shader_info.sampler(0, 'FLOAT_2D', 'inputTexture')
            shader_info.push_constant('FLOAT', 'opacity')
            
            shader_info.fragment_out(0, 'VEC4', 'fragColor')
            
            shader_info.vertex_source(vert_source)
            shader_info.fragment_source(frag_source)
            
            self.gpu_shader = gpu.shader.create_from_info(shader_info)
            return True
        except Exception:
            return False
    def create_fullscreen_quad(self):
        try:
            vertices = [
                (-1, -1),
                ( 1, -1),
                ( 1,  1),
                (-1,  1),
            ]
            
            texcoords = [
                (0, 0),
                (1, 0),
                (1, 1),
                (0, 1),
            ]
            
            indices = [
                (0, 1, 2),
                (0, 2, 3),
            ]
            
            self.batch = batch_for_shader(
                self.gpu_shader, 
                'TRIS',
                {
                    "position": vertices,
                    "texCoord_0": texcoords,
                },
                indices=indices
            )
            return True
        except Exception:
            return False
    def update_mouse_position(self, mouse_x, mouse_y):
        self.mouse_pos[0] = max(0.0, min(1.0, mouse_x))
        self.mouse_pos[1] = max(0.0, min(1.0, 1.0 - mouse_y))
        self.write_mouse_buffer()
    def update_region_size(self, width, height):
        w = max(1, int(width))
        h = max(1, int(height))
        old_region_size = self.region_size
        self.region_size = (w, h)
        
        size_changed = old_region_size != self.region_size
        
        if size_changed:
            # Recompute layout with new viewport size
            updated_container_data = parser_op.recompute_layout((w, h))
            
            if updated_container_data:
                self.container_data = updated_container_data
                
                # Rebuild container buffer with new layout
                container_array = []
                for i, container in enumerate(self.container_data):
                    container_struct = [
                        int(container.get('display', False)),
                        container.get('position', [0, 0])[0], container.get('position', [0, 0])[1],
                        container.get('size', [100, 100])[0], container.get('size', [100, 100])[1],
                        container.get('color', [1, 1, 1, 1])[0], container.get('color', [1, 1, 1, 1])[1], 
                        container.get('color', [1, 1, 1, 1])[2], container.get('color', [1, 1, 1, 1])[3],
                        container.get('color_1', [1, 1, 1, 1])[0], container.get('color_1', [1, 1, 1, 1])[1], 
                        container.get('color_1', [1, 1, 1, 1])[2], container.get('color_1', [1, 1, 1, 1])[3],
                        container.get('color_gradient_rot', 0.0),
                        container.get('hover_color', container_default.hover_color)[0], container.get('hover_color', container_default.hover_color)[1], 
                        container.get('hover_color', container_default.hover_color)[2], container.get('hover_color', container_default.hover_color)[3],
                        container.get('hover_color_1', container_default.hover_color_1)[0], container.get('hover_color_1', container_default.hover_color_1)[1], 
                        container.get('hover_color_1', container_default.hover_color_1)[2], container.get('hover_color_1', container_default.hover_color_1)[3],
                        container.get('hover_color_gradient_rot', 0.0),
                        container.get('click_color', container_default.click_color)[0], container.get('click_color', container_default.click_color)[1], 
                        container.get('click_color', container_default.click_color)[2], container.get('click_color', container_default.click_color)[3],
                        container.get('click_color_1', container_default.click_color_1)[0], container.get('click_color_1', container_default.click_color_1)[1], 
                        container.get('click_color_1', container_default.click_color_1)[2], container.get('click_color_1', container_default.click_color_1)[3],
                        container.get('click_color_gradient_rot', 0.0),
                        container.get('border_color', [1, 1, 1, 1])[0], container.get('border_color', [1, 1, 1, 1])[1], 
                        container.get('border_color', [1, 1, 1, 1])[2], container.get('border_color', [1, 1, 1, 1])[3],
                        container.get('border_color_1', [1, 1, 1, 1])[0], container.get('border_color_1', [1, 1, 1, 1])[1], 
                        container.get('border_color_1', [1, 1, 1, 1])[2], container.get('border_color_1', [1, 1, 1, 1])[3],
                        container.get('border_color_gradient_rot', 0.0),
                        container.get('border_radius', 0.0),
                        container.get('border_width', 0.0),
                        container.get('parent', -1),
                        int(container.get('overflow', False)),
                        container.get('box_shadow_offset', [0, 0, 0])[0], container.get('box_shadow_offset', [0, 0, 0])[1], 
                        container.get('box_shadow_offset', [0, 0, 0])[2],
                        container.get('box_shadow_blur', 0.0),
                        container.get('box_shadow_color', [0, 0, 0, 0])[0], container.get('box_shadow_color', [0, 0, 0, 0])[1], 
                        container.get('box_shadow_color', [0, 0, 0, 0])[2], container.get('box_shadow_color', [0, 0, 0, 0])[3],
                        int(container.get('passive', False))
                    ]
                    container_array.extend(container_struct)
                
                if self.container_buffer:
                    container_data_np = np.array(container_array, dtype=np.float32)
                    self.container_buffer.write(container_data_np.tobytes())
        
        if self.viewport_buffer:
            viewport_data = np.array([w, h, len(self.container_data)], dtype=np.float32)
            self.viewport_buffer.write(viewport_data.tobytes())
        
        if size_changed and self.output_texture:
            if self.blender_texture:
                self.blender_texture = None
            
            if self._safe_release_moderngl_object(self.output_texture):
                self.texture_size = self.region_size
                self.output_texture = self.mgl_context.texture(
                    self.texture_size,
                    4
                )
                self.output_texture.filter = (mgl.NEAREST, mgl.NEAREST)
                self.needs_texture_update = True
        
        return size_changed
    def update_click_value(self, value):
        self.click_value = value
        self.write_mouse_buffer()
    def on_scroll(self, delta, absolute_value):
        self.write_mouse_buffer()
    def on_mouse_event(self, event_type, data):
        if event_type == 'mouse':
            self.mouse_pos[0] = max(0.0, min(1.0, (data[0] + 1.0) / 2.0))
            self.mouse_pos[1] = max(0.0, min(1.0, (data[1] + 1.0) / 2.0))
        elif event_type == 'click':
            self.click_value = 1.0 if data else 0.0
        self.write_mouse_buffer()
    def write_mouse_buffer(self):
        if not self.mouse_buffer:
            return
        current_time = time.time() - self.start_time
        scroll_value = float(scroll_state.scroll_value)
        mouse_data = np.array([
            self.mouse_pos[0],
            self.mouse_pos[1],
            current_time,
            scroll_value,
            self.click_value,
            0.0
        ], dtype=np.float32)
        self.mouse_buffer.write(mouse_data.tobytes())
    
    def update_fps(self):
        current_time = time.perf_counter()
        frame_time = current_time - self.last_frame_time
        self.last_frame_time = current_time
        
        self.frame_times.append(frame_time)
        if len(self.frame_times) > 60:
            self.frame_times.pop(0)
        
        if len(self.frame_times) > 0:
            avg_frame_time = sum(self.frame_times) / len(self.frame_times)
            self.compute_fps = 1.0 / avg_frame_time if avg_frame_time > 0 else 0.0
    
    def check_if_changed(self):
        """Check if texture needs updating and update state. Called from modal loop."""
        changed = False
        
        if (abs(self.mouse_pos[0] - self.last_mouse_pos[0]) > 0.001 or 
            abs(self.mouse_pos[1] - self.last_mouse_pos[1]) > 0.001):
            self.last_mouse_pos = self.mouse_pos.copy()
            changed = True
        
        if self.click_value != self.last_click_value:
            self.last_click_value = self.click_value
            changed = True
        
        current_scroll = float(scroll_state.scroll_value)
        if abs(current_scroll - self.last_scroll_value) > 0.001:
            self.last_scroll_value = current_scroll
            changed = True
        
        if self.needs_texture_update:
            self.needs_texture_update = False
            changed = True
        
        self.texture_needs_readback = changed
        
        return changed
    
    def has_texture_changed(self):
        """Check if texture needs readback. Called from draw_texture."""
        return self.texture_needs_readback
    def run_compute_shader(self):
        if not (self.compute_shader and self.mouse_buffer and self.container_buffer and 
                self.viewport_buffer and self.output_texture):
            return False
            
        try:
            self.mouse_buffer.bind_to_storage_buffer(0)
            self.container_buffer.bind_to_storage_buffer(1)
            self.viewport_buffer.bind_to_storage_buffer(2)
            self.output_texture.bind_to_image(4, read=False, write=True)
            
            groups_x = (self.texture_size[0] + 15) // 16
            groups_y = (self.texture_size[1] + 15) // 16

            self.compute_shader.run(groups_x, groups_y, 1)
            
            return True
        except Exception:
            return False
    def initialize(self):
        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        self.region_size = (region.width, region.height)
                        break
                break
        
        if not self.load_container_data():
            return False
        if not self.init_moderngl_context():
            return False
        if not self.create_compute_shader():
            return False
        if not self.create_buffers_and_textures():
            return False
        if not self.create_blender_gpu_shader():
            return False
        if not self.create_fullscreen_quad():
            return False

        scroll_state.register_callback(self.on_scroll)
        self.scroll_callback_registered = True
        
        mouse_state.register_callback(self.on_mouse_event)
        self.mouse_callback_registered = True
        
        self.running = True
        self.write_mouse_buffer()
        
        self.needs_texture_update = True
        
        self.add_drawing_callback()
        return True
    def add_drawing_callback(self):
        self.draw_handler = bpy.types.SpaceView3D.draw_handler_add(
            self.draw_texture, (), 'WINDOW', 'POST_PIXEL'
        )
    def draw_texture(self):
        if not (self.running and self.gpu_shader and self.batch and self.output_texture):
            return
            
        try:
            if not self.has_texture_changed():
                if self.blender_texture:
                    gpu.state.blend_set('ALPHA')
                    gpu.state.depth_test_set('NONE')
                    
                    self.gpu_shader.bind()
                    self.gpu_shader.uniform_sampler("inputTexture", self.blender_texture)
                    self.gpu_shader.uniform_float("opacity", 1.0)
                    
                    gpu.matrix.push()
                    gpu.matrix.load_identity()
                    
                    self.batch.draw(self.gpu_shader)
                    gpu.matrix.pop()
                    
                    gpu.state.blend_set('NONE')
                    gpu.state.depth_test_set('LESS_EQUAL')
                return
            
            texture_data = self.output_texture.read()
            
            float_data = np.frombuffer(texture_data, dtype=np.uint8).astype(np.float32) / 255.0
            
            buffer = gpu.types.Buffer('FLOAT', len(float_data), float_data)
            
            if self.blender_texture:
                del self.blender_texture
            
            self.blender_texture = gpu.types.GPUTexture(
                self.texture_size,
                format = 'RGBA8',
                data   = buffer
            )
            
            gpu.state.blend_set('ALPHA')
            gpu.state.depth_test_set('NONE')
            
            self.gpu_shader.bind()
            self.gpu_shader.uniform_sampler("inputTexture", self.blender_texture)
            self.gpu_shader.uniform_float("opacity", 1.0)
            
            gpu.matrix.push()
            gpu.matrix.load_identity()
            
            self.batch.draw(self.gpu_shader)
            gpu.matrix.pop()
            
            gpu.state.blend_set('NONE')
            gpu.state.depth_test_set('LESS_EQUAL')
            
        except Exception:
            pass
    
    def cleanup(self):
        self.running = False
        
        if self.draw_handler:
            bpy.types.SpaceView3D.draw_handler_remove(self.draw_handler, 'WINDOW')
            self.draw_handler = None
        
        if self.blender_texture:
            self.blender_texture = None
        
        self.needs_texture_update = True
        self.last_mouse_pos = [0.5, 0.5]
        self.last_click_value = 0.0
        self.last_scroll_value = 0.0
        
        if self._safe_release_moderngl_object(self.mouse_buffer):
            self.mouse_buffer = None
        if self._safe_release_moderngl_object(self.container_buffer):
            self.container_buffer = None
        if self._safe_release_moderngl_object(self.viewport_buffer):
            self.viewport_buffer = None
        if self._safe_release_moderngl_object(self.output_texture):
            self.output_texture = None
        if self._safe_release_moderngl_object(self.compute_shader):
            self.compute_shader = None
        
        if self.mgl_context:
            try:
                self.mgl_context.gc()
            except AttributeError as e:
                if "'InvalidObject' object has no attribute 'release'" not in str(e):
                    pass
            except Exception:
                pass
            finally:
                self.mgl_context = None
        
        try:
            import gc
            gc.collect()
        except:
            pass
        
        if self.scroll_callback_registered:
            scroll_state.unregister_callback(self.on_scroll)
            self.scroll_callback_registered = False
        
        if self.mouse_callback_registered:
            mouse_state.unregister_callback(self.on_mouse_event)
            self.mouse_callback_registered = False
    def update_container_buffer_full(self, hit_container_data):
        """Update entire container buffer with current interaction states"""
        if not self.container_buffer or not hit_container_data:
            return False
        
        try:
            container_array = []
            updates_made = 0
            
            for i, container in enumerate(hit_container_data):
                state_changed = (
                    container.get('_hovered', False) != container.get('_prev_hovered', False) or
                    container.get('_clicked', False) != container.get('_prev_clicked', False)
                )
                
                if state_changed:
                    updates_made += 1
                
                current_color = container.get('color', [1, 1, 1, 1]).copy()
                current_color_1 = container.get('color_1', [1, 1, 1, 1]).copy()
                
                container_struct = [
                    int(container.get('display', False)),
                    container.get('position', [0, 0])[0], container.get('position', [0, 0])[1],
                    container.get('size', [100, 100])[0], container.get('size', [100, 100])[1],
                    current_color[0], current_color[1], current_color[2], current_color[3],
                    current_color_1[0], current_color_1[1], current_color_1[2], current_color_1[3],
                    container.get('color_gradient_rot', 0.0),
                    container.get('hover_color', container_default.hover_color)[0], container.get('hover_color', container_default.hover_color)[1], 
                    container.get('hover_color', container_default.hover_color)[2], container.get('hover_color', container_default.hover_color)[3],
                    container.get('hover_color_1', container_default.hover_color_1)[0], container.get('hover_color_1', container_default.hover_color_1)[1], 
                    container.get('hover_color_1', container_default.hover_color_1)[2], container.get('hover_color_1', container_default.hover_color_1)[3],
                    container.get('hover_color_gradient_rot', 0.0),
                    container.get('click_color', container_default.click_color)[0], container.get('click_color', container_default.click_color)[1], 
                    container.get('click_color', container_default.click_color)[2], container.get('click_color', container_default.click_color)[3],
                    container.get('click_color_1', container_default.click_color_1)[0], container.get('click_color_1', container_default.click_color_1)[1], 
                    container.get('click_color_1', container_default.click_color_1)[2], container.get('click_color_1', container_default.click_color_1)[3],
                    container.get('click_color_gradient_rot', 0.0),
                    container.get('border_color', [1, 1, 1, 1])[0], container.get('border_color', [1, 1, 1, 1])[1], 
                    container.get('border_color', [1, 1, 1, 1])[2], container.get('border_color', [1, 1, 1, 1])[3],
                    container.get('border_color_1', [1, 1, 1, 1])[0], container.get('border_color_1', [1, 1, 1, 1])[1], 
                    container.get('border_color_1', [1, 1, 1, 1])[2], container.get('border_color_1', [1, 1, 1, 1])[3],
                    container.get('border_color_gradient_rot', 0.0),
                    container.get('border_radius', 0.0),
                    container.get('border_width', 0.0),
                    container.get('parent', -1),
                    int(container.get('overflow', False)),
                    container.get('box_shadow_offset', [0, 0, 0])[0], container.get('box_shadow_offset', [0, 0, 0])[1], 
                    container.get('box_shadow_offset', [0, 0, 0])[2],
                    container.get('box_shadow_blur', 0.0),
                    container.get('box_shadow_color', [0, 0, 0, 0])[0], container.get('box_shadow_color', [0, 0, 0, 0])[1], 
                    container.get('box_shadow_color', [0, 0, 0, 0])[2], container.get('box_shadow_color', [0, 0, 0, 0])[3],
                    int(container.get('passive', False))
                ]
                container_array.extend(container_struct)
            
            # Update entire buffer
            container_data_np = np.array(container_array, dtype=np.float32)
            self.container_buffer.write(container_data_np.tobytes())
            
            if updates_made > 0:
                self.needs_texture_update = True
            
            return True
        except Exception:
            return False

class XWZ_OT_start_ui(Operator):
    bl_idname      = "xwz.start_ui"
    bl_label       = "Start puree"
    bl_description = "Start puree UI"
    
    def execute(self, context):
        global _render_data, _modal_timer
        
        if _render_data and _render_data.running:
            self.report({'WARNING'}, "Demo already running")
            return {'CANCELLED'}
        
        _render_data = RenderPipeline()
        
        if not _render_data.initialize():
            self.report({'ERROR'}, "Failed to initialize compute shader demo")
            _render_data = None
            return {'CANCELLED'}

        # Start native-optimized hit detection
        try:
            bpy.ops.xwz.hit_detect('INVOKE_DEFAULT')
        except Exception as e:
            self.report({'WARNING'}, f"Failed to start hit detect modal: {e}")

        try:
            bpy.ops.xwz.scroll_modal_launch('INVOKE_DEFAULT')
        except Exception as e:
            self.report({'WARNING'}, f"Failed to start scroll modal: {e}")
        
        try:
            bpy.ops.xwz.mouse_modal_launch('INVOKE_DEFAULT')
        except Exception as e:
            self.report({'WARNING'}, f"Failed to start mouse modal: {e}")
        
        context.window_manager.modal_handler_add(self)
        _modal_timer = context.window_manager.event_timer_add(0.016, window=context.window)
        
        for _container_id in parser_op.image_blocks:
            block = parser_op.image_blocks[_container_id]
            bpy.ops.xwz.draw_image(
                container_id = _container_id,
                image_name   = block['image_name'],
                x_pos        = block['x_pos'],
                y_pos        = block['y_pos'],
                width        = block['width'],
                height       = block['height'],
                mask_x       = block['mask_x'],
                mask_y       = block['mask_y'],
                mask_width   = block['mask_width'],
                mask_height  = block['mask_height'],
                aspect_ratio = block['aspect_ratio'],
                align_h      = block.get('align_h', 'LEFT').upper(),
                align_v      = block.get('align_v', 'TOP').upper()
            )
        
        for _container_id in parser_op.text_blocks:
            block = parser_op.text_blocks[_container_id]
            bpy.ops.xwz.draw_text(
                container_id = _container_id,
                text          = block['text'],
                font_name     = block['font'],
                size          = block['text_scale'],
                x_pos         = block['text_x'],
                y_pos         = block['text_y'],
                color         = block['text_color'],
                mask_x        = block['mask_x'],
                mask_y        = block['mask_y'],
                mask_width    = block['mask_width'],
                mask_height   = block['mask_height'],
                align_h       = block.get('align_h', 'LEFT').upper(),
                align_v       = block.get('align_v', 'CENTER').upper()
            )
        
        for _container_id in parser_op.text_input_blocks:
            block = parser_op.text_input_blocks[_container_id]
            bpy.ops.xwz.create_text_input(
                container_id = _container_id,
                placeholder  = block['placeholder'],
                font_name    = block['font'],
                size         = block['text_scale'],
                x_pos        = block['x_pos'],
                y_pos        = block['y_pos'],
                color        = block['text_color'],
                mask_x       = block['mask_x'],
                mask_y       = block['mask_y'],
                mask_width   = block['mask_width'],
                mask_height  = block['mask_height'],
                align_h      = block.get('align_h', 'LEFT').upper(),
                align_v      = block.get('align_v', 'TOP').upper()
            )

        self.report({'INFO'}, "UI Started")
        return {'RUNNING_MODAL'}
    
    def modal(self, context, event):
        global _render_data
        
        if not (_render_data and _render_data.running):
            self.cancel(context)
            return {'CANCELLED'}
        
        # Handle window deactivate to catch resize events
        if event.type == 'WINDOW_DEACTIVATE':
            area = context.area
            region = context.region
            
            if area and region:
                size_changed = _render_data.update_region_size(region.width, region.height)
                if size_changed:
                    from .hit_op import _container_data
                    if _container_data:
                        _render_data.update_container_buffer_full(_container_data)
                    
                    _render_data.run_compute_shader()
                    
                    for area in context.screen.areas:
                        if area.type == 'VIEW_3D':
                            area.tag_redraw()
        
        if event.type == 'TIMER':
            area = context.area
            region = context.region
            
            if area and region:
                _render_data.update_fps()
                
                size_changed = _render_data.update_region_size(region.width, region.height)

                texture_changed = _render_data.check_if_changed()
                
                # Check if script callbacks modified container state
                state_synced = parser_op.sync_dirty_containers()
                if state_synced:
                    # Container state changed via scripts, need to update
                    from . import hit_op
                    from . import text_op
                    new_data = parser_op._container_json_data
                    old_data = hit_op._container_data
                    
                    if old_data and len(old_data) == len(new_data):
                        for i in range(len(new_data)):
                            runtime_keys = ['_hovered', '_prev_hovered', '_clicked', '_prev_clicked', 
                                          '_toggled', '_prev_toggled', '_toggle_value', '_scroll_value']
                            for key in runtime_keys:
                                if key in old_data[i]:
                                    new_data[i][key] = old_data[i][key]
                    
                    hit_op._container_data = new_data
                    
                    # Update text instances with new text content
                    for text_instance in text_op._text_instances:
                        container_id = text_instance.container_id
                        if container_id in parser_op.text_blocks:
                            block = parser_op.text_blocks[container_id]
                            text_instance.update_all(
                                text=block['text'],
                                font_name=block['font'],
                                size=block['text_scale'],
                                pos=[block['text_x'], block['text_y']],
                                color=block['text_color'],
                                mask=[block['mask_x'], block['mask_y'], block['mask_width'], block['mask_height']],
                                align_h=block.get('align_h', 'LEFT').upper(),
                                align_v=block.get('align_v', 'CENTER').upper()
                            )
                    
                    # Update text input instances with new layout
                    from . import text_input_op
                    for input_instance in text_input_op._text_input_instances:
                        container_id = input_instance.container_id
                        if container_id in parser_op.text_input_blocks:
                            block = parser_op.text_input_blocks[container_id]
                            bpy.ops.xwz.update_text_input(
                                instance_id=input_instance.id,
                                placeholder=block['placeholder'],
                                font_name=block['font'],
                                size=block['text_scale'],
                                x_pos=block['x_pos'],
                                y_pos=block['y_pos'],
                                color=block['text_color'],
                                mask_x=block['mask_x'],
                                mask_y=block['mask_y'],
                                mask_width=block['mask_width'],
                                mask_height=block['mask_height'],
                                align_h=block.get('align_h', 'LEFT').upper(),
                                align_v=block.get('align_v', 'TOP').upper()
                            )
                    
                    # Update image instances with new image content
                    from . import img_op
                    for image_instance in img_op._image_instances:
                        container_id = image_instance.container_id
                        if container_id in parser_op.image_blocks:
                            block = parser_op.image_blocks[container_id]
                            image_instance.update_all(
                                image_name=block['image_name'],
                                pos=[block['x_pos'], block['y_pos']],
                                size=[block['width'], block['height']],
                                mask=[block['mask_x'], block['mask_y'], block['mask_width'], block['mask_height']],
                                aspect_ratio=block['aspect_ratio'],
                                align_h=block.get('align_h', 'LEFT').upper(),
                                align_v=block.get('align_v', 'TOP').upper()
                            )
                    
                    texture_changed = True
                
                # Run compute shader if viewport resized or other state changed
                if texture_changed or size_changed:
                    # If size changed, update hit_op's container data reference
                    if size_changed:
                        from . import hit_op
                        new_data = parser_op._container_json_data
                        old_data = hit_op._container_data
                        
                        if old_data and len(old_data) == len(new_data):
                            for i in range(len(new_data)):
                                runtime_keys = ['_hovered', '_prev_hovered', '_clicked', '_prev_clicked', 
                                              '_toggled', '_prev_toggled', '_toggle_value', '_scroll_value']
                                for key in runtime_keys:
                                    if key in old_data[i]:
                                        new_data[i][key] = old_data[i][key]
                        
                        hit_op._container_data = new_data
                    
                    from .hit_op import _container_data
                    if _container_data:
                        _render_data.update_container_buffer_full(_container_data)
                    
                    _render_data.run_compute_shader()
            
            for area in context.screen.areas:
                if area.type == 'VIEW_3D':
                    area.tag_redraw()

        elif event.type in {'ESC'}:
            self.cancel(context)
            return {'CANCELLED'}
        
        return {'PASS_THROUGH'}
    
    def cancel(self, context):
        global _render_data, _modal_timer
        
        if _modal_timer:
            context.window_manager.event_timer_remove(_modal_timer)
            _modal_timer = None
        
        if _render_data:
            _render_data.cleanup()
            _render_data = None
        
        bpy.ops.xwz.hit_stop()
        scroll_state.stop_scrolling()
        mouse_state.stop_mouse_tracking()

class XWZ_OT_stop_ui(Operator):
    bl_idname      = "xwz.stop_ui"
    bl_label       = "Stop puree"
    bl_description = "Stop puree UI"
    
    def execute(self, context):
        global _render_data, _modal_timer
        
        if _modal_timer:
            context.window_manager.event_timer_remove(_modal_timer)
            _modal_timer = None
        
        if _render_data:
            _render_data.cleanup()
            _render_data = None

        bpy.ops.xwz.hit_stop()
        scroll_state.stop_scrolling()
        mouse_state.stop_mouse_tracking()
        
        try:
            bpy.ops.xwz.clear_text()
            bpy.ops.xwz.clear_text_inputs()
            bpy.ops.xwz.clear_images()
        except Exception:
            pass

        for area in context.screen.areas:
            if area.type == 'VIEW_3D':
                area.tag_redraw()
            
        self.report({'INFO'}, "Compute shader demo stopped")
        return {'FINISHED'}

classes = [
    XWZ_OT_start_ui,
    XWZ_OT_stop_ui, 
    XWZ_OT_scroll,
    XWZ_OT_scroll_launch,
    XWZ_OT_mouse,
    XWZ_OT_mouse_launch,
    XWZ_OT_ui_parser
]

def register():
    for cls in classes:
        bpy.utils.register_class(cls)

def unregister():
    global _render_data, _modal_timer
    
    if _modal_timer:
        try:
            context = bpy.context
            context.window_manager.event_timer_remove(_modal_timer)
        except:
            pass
        _modal_timer = None
    
    if _render_data:
        _render_data.cleanup()
        _render_data = None
    
    scroll_state.stop_scrolling()
    mouse_state.stop_mouse_tracking()
    
    try:
        import gc
        import sys
        
        gc.collect()
        
        modules_to_remove = [name for name in sys.modules.keys() if name.startswith('moderngl')]
        for module_name in modules_to_remove:
            if module_name in sys.modules:
                try:
                    del sys.modules[module_name]
                except:
                    pass
        
        gc.collect()
        
    except Exception:
        pass
    
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)


import os
import re
import yaml
import math

from stretchable import Node
from stretchable.style import PCT, AUTO, PT
from stretchable import Edge
from stretchable.style.props import BoxSizing
from stretchable.style.props import FlexDirection
from stretchable.style.props import AlignItems, JustifyContent
from stretchable.style.props import Display, Position
from stretchable.style.geometry.rect import RectPointsPercent
from stretchable.style.geometry.length import LengthPointsPercent
from tinycss2 import ast
from tinycss2 import parse_stylesheet, parse_declaration_list
from textual.color import Color

from .components.container import Container
from .components.style import Style
from .scss_compiler import SCSSCompiler

node_flat = {}
node_flat_abs = {}

def gamma_correct(value):
    return math.pow(value, 2.2)

def apply_gamma_correction(r, g, b, a):
    return [
        gamma_correct(r),
        gamma_correct(g),
        gamma_correct(b),
        a 
    ]

class CSSParser:
    def __init__(self):
        self.styles = {}
        self.variables = {}
    
    def _extract_variables(self, styles):
        variables = {}
        for selector, declarations in styles.items():
            for prop, value in declarations.items():
                if prop.startswith('--'):
                    variables[prop] = value
        return variables
    
    def _resolve_variables(self, value, variables, visited=None):
        import re
        if visited is None:
            visited = set()
        
        var_pattern = r'var\(([^)]+)\)'
        
        def replace_var(match):
            var_name = match.group(1).strip()
            if var_name in visited:
                return match.group(0)
            if var_name in variables:
                new_visited = visited.copy()
                new_visited.add(var_name)
                resolved_value = variables[var_name]
                result = self._resolve_variables(resolved_value, variables, new_visited)
                return result
            return match.group(0)
        
        resolved = re.sub(var_pattern, replace_var, value)
        resolved = ' '.join(resolved.split())
        return resolved
    
    def parse(self, css_string):
        self.styles = {}
        rules = parse_stylesheet(css_string)
        
        for rule in rules:
            if hasattr(rule, 'content'):
                selector = ''.join(token.serialize() for token in rule.prelude).strip()
                declarations = {}
                
                content_str = ''.join(token.serialize() for token in rule.content)
                decl_list = parse_declaration_list(content_str)
                
                for decl in decl_list:
                    if isinstance(decl, ast.Declaration):
                        value = ''.join(token.serialize() for token in decl.value).strip()
                        declarations[decl.name] = value
                
                self.styles[selector] = declarations
        
        self.variables = self._extract_variables(self.styles)
        
        for selector, declarations in self.styles.items():
            for prop, value in declarations.items():
                if not prop.startswith('--'):
                    self.styles[selector][prop] = self._resolve_variables(value, self.variables)
        
        return self.styles

class Settings():
    def __init__(self):
        self.scroll_speed = 0

class Styles():
    def __init__(self):
        pass

class Theme():
    def __init__(self):
        self.name         = ""
        self.author       = ""
        self.version      = ""
        self.scripts      = []
        self.style_files  = []
        self.default_font = ""
        self.components   = ""
        self.palette      = {}
        self.styles       = Styles()
        self.root         = Container()
        
class UI():
    def __init__(self, path=None, base_dir=None, canvas_size=(800, 600)):
        self.selected_theme = "xwz_default"
        self.settings       = Settings()
        self.theme          = Theme()
        self.json_data      = []
        self.abs_json_data  = []
        self.root_node      = None
        self.canvas_size    = canvas_size

        self.parse_toml(path, base_dir)
        self.parse_css()
        self.create_node_tree(canvas_size)
        self.flatten_node_tree()

    def get_by_id(self, target_id):
        return self.theme.root.get_by_id(target_id)

    def load_conf_file(self, path):
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
        return data
    
    def parse_toml(self, path=None, base_dir=None):
        data     = self.load_conf_file(path)
        ui_data  = data.get('app', {})
        theme    = ui_data['theme']

        self.selected_theme        = ui_data['selected_theme']
        self.default_theme         = ui_data['default_theme']

        self.theme_index = -1
        for _theme_ in theme:
            if _theme_['name'] == self.selected_theme:
                self.theme_index = theme.index(_theme_)
                break
        if self.theme_index == -1:
            for _theme_ in theme:
                if _theme_['name'] == self.default_theme:
                    self.theme_index = theme.index(_theme_)
                    break
        if self.theme_index == -1:
            self.theme_index   = 0
            self.default_theme = theme[0]['name']

        root = ui_data['theme'][self.theme_index]['root']

        self.theme.name         = theme[self.theme_index]['name']
        self.theme.author       = theme[self.theme_index]['author']
        self.theme.version      = theme[self.theme_index]['version']
        self.theme.scripts      = theme[self.theme_index]['scripts']
        self.theme.style_files  = theme[self.theme_index]['styles']
        self.theme.default_font = theme[self.theme_index]['default_font']
        self.theme.components   = theme[self.theme_index]['components']

        def load_container(container_data, parent_container):
            for attr_name, attr_value in container_data.items():

                if isinstance(attr_value, dict):
                    has_component_data = 'data' in attr_value and isinstance(attr_value['data'], str) and attr_value['data'].startswith('[') and attr_value['data'].endswith(']')
                    
                    child_container = Container()
                    if parent_container.id == "root":
                        child_container.id = attr_name
                    else:
                        child_container.id = f"{parent_container.id}_{attr_name}"
                    child_container.parent = parent_container
                    parent_container.children.append(child_container)
                    
                    for child_attr_name, child_attr_value in attr_value.items():
                        if not isinstance(child_attr_value, dict):
                            if hasattr(child_container, child_attr_name):
                                if not (child_attr_name == 'data' and has_component_data):
                                    setattr(child_container, child_attr_name.replace('-', '_'), child_attr_value)
                    
                    if has_component_data:
                        component_ref = attr_value['data']
                        component_dir = os.path.join(base_dir, self.theme.components)
                        component_loaded = False
                        
                        component_params = {}
                        for param_name, param_value in attr_value.items():
                            if not isinstance(param_value, dict) and param_name != 'data':
                                component_params[param_name] = param_value
                        
                        for root, dirs, files in os.walk(component_dir):
                            for filename in files:
                                if filename.endswith('.yaml') and f'[{filename.replace(".yaml", "")}]' == component_ref:
                                    file_path = os.path.join(root, filename)
                                    component_base_name = filename.replace('.yaml', '')
                                    scss_file_path = os.path.join(root, f"{component_base_name}.scss")
                                    
                                    with open(file_path, 'r') as f:
                                        component_data = yaml.safe_load(f)
                                        component_key = component_ref.replace("[",'').replace("]",'')
                                        
                                        if os.path.exists(scss_file_path):
                                            scss_compiler = SCSSCompiler()
                                            namespace = child_container.id
                                            compiled_css = scss_compiler.compile_file(
                                                scss_file_path,
                                                namespace=namespace,
                                                param_overrides=component_params,
                                                component_name=component_base_name
                                            )
                                            
                                            css_parser = CSSParser()
                                            component_styles = css_parser.parse(compiled_css)
                                            for selector, declarations in component_styles.items():
                                                style_obj = Style()
                                                style_obj.id = selector
                                                for prop, value in declarations.items():
                                                    attr_name_parsed, attr_value_parsed = self.parse_container_props_from_style(prop, value)
                                                    setattr(style_obj, attr_name_parsed, attr_value_parsed)
                                                self.theme.styles.__dict__[selector] = style_obj
                                        
                                        def substitute_params(value, params):
                                            if not isinstance(value, str):
                                                return value
                                            
                                            pattern = r'\{\{(\w+)\s*,\s*["\']([^"\']*?)["\']\}\}'
                                            
                                            def replace_param(match):
                                                param_name = match.group(1)
                                                default_value = match.group(2)
                                                return str(params.get(param_name, default_value))
                                            
                                            return re.sub(pattern, replace_param, value)
                                        
                                        def load_component_with_namespace(comp_data, parent, namespace_prefix, params):
                                            for attr_name, attr_value in comp_data.items():
                                                if isinstance(attr_value, dict):
                                                    namespaced_child = Container()
                                                    namespaced_child.id = f"{parent.id}_{attr_name}"
                                                    namespaced_child.parent = parent
                                                    parent.children.append(namespaced_child)
                                                    
                                                    for child_attr_name, child_attr_value in attr_value.items():
                                                        if not isinstance(child_attr_value, dict):
                                                            if hasattr(namespaced_child, child_attr_name):
                                                                substituted_value = substitute_params(child_attr_value, params)
                                                                if child_attr_name == 'style' and isinstance(substituted_value, str):
                                                                    if substituted_value == component_base_name:
                                                                        substituted_value = child_container.id
                                                                    elif substituted_value.startswith(component_base_name + '_'):
                                                                        substituted_value = substituted_value.replace(component_base_name, child_container.id, 1)
                                                                setattr(namespaced_child, child_attr_name.replace('-', '_'), substituted_value)
                                                    
                                                    load_component_with_namespace(attr_value, namespaced_child, namespaced_child.id, params)
                                                else:
                                                    if hasattr(parent, attr_name):
                                                        substituted_value = substitute_params(attr_value, params)
                                                        if attr_name == 'style' and isinstance(substituted_value, str):
                                                            if substituted_value == component_base_name:
                                                                substituted_value = child_container.id
                                                            elif substituted_value.startswith(component_base_name + '_'):
                                                                substituted_value = substituted_value.replace(component_base_name, child_container.id, 1)
                                                        setattr(parent, attr_name.replace('-', '_'), substituted_value)
                                        
                                        load_component_with_namespace(component_data[component_key], child_container, attr_name, component_params)
                                        component_loaded = True
                                    break
                            if component_loaded:
                                break
                    else:
                        load_container(attr_value, child_container)

                else:
                    if hasattr(parent_container, attr_name):
                        setattr(parent_container, attr_name.replace('-', '_'), attr_value)


        self.theme.root.id = "root"
        load_container(root, self.theme.root)

    def parse_container_props_from_style(self, attr_name, attr_value):
        attr_name = attr_name.replace('-', '_')

        if attr_name.startswith('__'):
                    attr_name = attr_name[1:]

        attr_name = attr_name.replace('background_color', 'color')

        color_props = [
            'color', 'color_1',
            'hover_color', 'hover_color_1',
            'click_color', 'click_color_1',
            'border_color', 'border_color_1',
            'text_color', 'text_color_1',
            'box_shadow_color'
            ]
        
        float_props = [
            'border_radius', 'border_width',
            'text_scale', 'text_x', 'text_y',
            'box_shadow_blur'
            ]

        rotation_props = [
            'color_gradient_rot',
            'hover_color_gradient_rot',
            'click_color_gradient_rot',
            'border_color_gradient_rot',
            'text_color_gradient_rot'
            ]

        bool_props = [
            'aspect_ratio'
            ]
        
        string_props = [
            'overflow', 'display', 'position',
            'flex_direction', 'align_items', 'justify_content'
            ]

        if attr_name in color_props:
            value_color = Color.parse(attr_value)
            srgb_normalized = [value_color.r / 255.0, value_color.g / 255.0, value_color.b / 255.0, value_color.a]
            attr_value = apply_gamma_correction(*srgb_normalized)

        elif attr_name in float_props:
            attr_value = float(attr_value.replace('px', '').strip())

        elif attr_name in rotation_props:
            attr_value = float(attr_value.replace('deg', '').strip())

        elif attr_name == 'box_shadow_offset':
            values = attr_value.strip().split()
            if len(values) == 2:
                x_offset = float(values[0].replace('px', '').strip())
                y_offset = float(values[1].replace('px', '').strip())
                attr_value = [x_offset, y_offset, 0.0]
            else:
                attr_value = [0.0, 0.0, 0.0]

        elif attr_name in bool_props:
            attr_value = attr_value.strip().lower() in ('true', '1', 'yes')
        
        elif attr_name in string_props:
            attr_value = attr_value.strip().upper().replace('-', '_')

        return attr_name, attr_value

    def parse_css(self):
        from . import get_addon_root
        addon_dir  = get_addon_root()
        style_str = ""
        scss_compiler = SCSSCompiler()
        
        for _style_file in self.theme.style_files:
            file_path = os.path.join(addon_dir, _style_file)
            if _style_file.endswith('.scss'):
                compiled_css = scss_compiler.compile_file(file_path)
                style_str += compiled_css
            elif _style_file.endswith('.css'):
                with open(file_path, 'r') as f:
                    style_str += f.read()
        
        css_string = style_str
        parser = CSSParser()
        styles = parser.parse(css_string)
        for selector, declarations in styles.items():
            style_obj = Style()
            selector_clean = selector.lstrip('.')
            style_obj.id = selector_clean
            for prop, value in declarations.items():
                attr_name, attr_value = self.parse_container_props_from_style(prop, value)
                setattr(style_obj, attr_name, attr_value)
            self.theme.styles.__dict__[selector_clean] = style_obj
        def apply_styles_to_containers(container):
            if hasattr(container, 'style') and container.style:
                style_name = container.style
                if isinstance(style_name, str):
                    if hasattr(self.theme.styles, style_name):
                        original_style = getattr(self.theme.styles, style_name)
                        style_copy = Style()
                        style_copy.id = original_style.id
                        for attr_name in dir(original_style):
                            if not attr_name.startswith('_'):
                                try:
                                    attr_value = getattr(original_style, attr_name)
                                    if not callable(attr_value):
                                        if isinstance(attr_value, list):
                                            setattr(style_copy, attr_name, attr_value.copy())
                                        else:
                                            setattr(style_copy, attr_name, attr_value)
                                except AttributeError:
                                    pass
                        container.style = style_copy
                    else:
                        print(f"Warning: Style '{style_name}' not found, using default")
                        default_style = Style()
                        setattr(default_style, 'width', "100%")
                        setattr(default_style, 'height', "100%")
                        container.style = default_style
            for child in container.children:
                apply_styles_to_containers(child)
        apply_styles_to_containers(self.theme.root)

    def create_node_tree(self, canvas_size=(800, 600)):
        def get_all_nodes(container, node):
            border_box     = node.get_box(Edge.BORDER, relative=True)
            border_box_abs = node.get_box(Edge.BORDER, relative=False)
            content_box    = node.get_box(Edge.CONTENT, relative=True)
            content_box_abs = node.get_box(Edge.CONTENT, relative=False)
            padding_box    = node.get_box(Edge.PADDING, relative=True)
            margin_box     = node.get_box(Edge.MARGIN, relative=True)
            margin_box_abs = node.get_box(Edge.MARGIN, relative=False)
            
            edge_used, edge_used_abs = border_box, border_box_abs

            node_flat[container.id] = {
                'x'      : edge_used.x,
                'y'      : edge_used.y,
                'width'  : edge_used.width,
                'height' : edge_used.height
            }

            node_flat_abs[container.id] = {
                'x'      : edge_used_abs.x,
                'y'      : edge_used_abs.y,
                'width'  : edge_used_abs.width,
                'height' : edge_used_abs.height
            }
            
            for i, _container in enumerate(container.children):
                get_all_nodes(_container, node[i])

            return
        def parse_css_value(value_str):
            value_str = str(value_str).lower()
            if 'px' in value_str and 'calc(' not in value_str:
                return LengthPointsPercent.from_any(int(value_str.replace('px', '')) * PT)
            if '%' in value_str and 'calc(' not in value_str:
                return LengthPointsPercent.from_any(int(value_str.replace('%', '')) * PCT)
            if 'auto' in value_str and 'calc(' not in value_str:
                return AUTO
            return LengthPointsPercent.from_any(0 * PT)
        def parse_padding_values(container):
            top = right = bottom = left = LengthPointsPercent.from_any(0 * PT)
            if hasattr(container.style, 'padding_top'):
                top = parse_css_value(container.style.padding_top)
            if hasattr(container.style, 'padding_right'):
                right = parse_css_value(container.style.padding_right)
            if hasattr(container.style, 'padding_bottom'):
                bottom = parse_css_value(container.style.padding_bottom)
            if hasattr(container.style, 'padding_left'):
                left = parse_css_value(container.style.padding_left)
            if hasattr(container.style, 'padding') and isinstance(container.style.padding, str):
                padding_str = container.style.padding.strip().lower()
                if 'calc(' not in padding_str:
                    # Split on whitespace - this handles multiple spaces correctly
                    values = padding_str.split()
                    if len(values) == 1:
                        val = parse_css_value(values[0])
                        top = right = bottom = left = val
                    elif len(values) == 2:
                        vertical   = parse_css_value(values[0])
                        horizontal = parse_css_value(values[1])
                        top        = bottom = vertical
                        right      = left   = horizontal
                    elif len(values) == 3:
                        top        = parse_css_value(values[0])
                        horizontal = parse_css_value(values[1])
                        bottom     = parse_css_value(values[2])
                        right      = left = horizontal
                    elif len(values) == 4:
                        top    = parse_css_value(values[0])
                        right  = parse_css_value(values[1])
                        bottom = parse_css_value(values[2])
                        left   = parse_css_value(values[3])
            return RectPointsPercent.from_any([top, right, bottom, left])
        def parse_margin_values(container):
            top = right = bottom = left = LengthPointsPercent.from_any(0 * PT)
            if hasattr(container.style, 'margin_top'):
                top = parse_css_value(container.style.margin_top)
            if hasattr(container.style, 'margin_right'):
                right = parse_css_value(container.style.margin_right)
            if hasattr(container.style, 'margin_bottom'):
                bottom = parse_css_value(container.style.margin_bottom)
            if hasattr(container.style, 'margin_left'):
                left = parse_css_value(container.style.margin_left)
            if hasattr(container.style, 'margin') and isinstance(container.style.margin, str):
                margin_str = container.style.margin.strip().lower()
                if 'calc(' not in margin_str:
                    # Split on whitespace - this handles multiple spaces correctly
                    values = margin_str.split()
                    
                    if len(values) == 1:
                        val = parse_css_value(values[0])
                        top = right = bottom = left = val
                    elif len(values) == 2:
                        vertical = parse_css_value(values[0])
                        horizontal = parse_css_value(values[1])
                        top = bottom = vertical
                        right = left = horizontal
                    elif len(values) == 3:
                        top = parse_css_value(values[0])
                        horizontal = parse_css_value(values[1])
                        bottom = parse_css_value(values[2])
                        right = left = horizontal
                    elif len(values) == 4:
                        top = parse_css_value(values[0])
                        right = parse_css_value(values[1])
                        bottom = parse_css_value(values[2])
                        left = parse_css_value(values[3])
            return RectPointsPercent.from_any([top, right, bottom, left])
        def parse_border_values(container):
            width_top = width_right = width_bottom = width_left = LengthPointsPercent.from_any(0 * PT)
            
            if hasattr(container.style, 'border_width') and isinstance(container.style.border_width, str):
                border_width_str = container.style.border_width.strip().lower()
                if 'calc(' not in border_width_str:
                    # Split on whitespace - this handles multiple spaces correctly
                    values = border_width_str.split()
                    
                    if len(values) == 1:
                        val = parse_css_value(values[0])
                        width_top = width_right = width_bottom = width_left = val
                    elif len(values) == 2:
                        vertical = parse_css_value(values[0])
                        horizontal = parse_css_value(values[1])
                        width_top = width_bottom = vertical
                        width_right = width_left = horizontal
                    elif len(values) == 3:
                        width_top = parse_css_value(values[0])
                        horizontal = parse_css_value(values[1])
                        width_bottom = parse_css_value(values[2])
                        width_right = width_left = horizontal
                    elif len(values) == 4:
                        width_top = parse_css_value(values[0])
                        width_right = parse_css_value(values[1])
                        width_bottom = parse_css_value(values[2])
                        width_left = parse_css_value(values[3])
            
            if hasattr(container.style, 'border') and isinstance(container.style.border, str):
                border_str = container.style.border.strip().lower()
                if 'calc(' not in border_str:
                    # Split on whitespace - this handles multiple spaces correctly
                    parts = border_str.split()
                    for part in parts:
                        if 'px' in part or '%' in part:
                            val = parse_css_value(part)
                            width_top = width_right = width_bottom = width_left = val
                        elif part.startswith('#') or part in ['red', 'blue', 'green', 'black', 'white', 'transparent']:
                            setattr(container.style, 'border_color_css', part)
            
            if hasattr(container.style, 'border_color') and isinstance(container.style.border_color, str):
                setattr(container.style, 'border_color_css', container.style.border_color.lower())
                        
            return RectPointsPercent.from_any([width_top, width_right, width_bottom, width_left])
        def create_node(container):
            if not hasattr(container, 'style') or container.style is None or isinstance(container.style, str):
                default_style = Style()
                setattr(default_style, 'width', "100%")
                setattr(default_style, 'height', "100%")
                container.style = default_style
            
            disp_str     = container.style.display.lower()
            pos_str      = container.style.position.lower()
            position_val = Position.RELATIVE
            display_val  = Display.FLEX

            width_val    = container.style.width
            height_val   = container.style.height
            width_pct    = 0
            height_pct   = 0
            
            flex_dir_str        = container.style.flex_direction.lower().replace('-', '_')
            align_str           = container.style.align_items.lower().replace('-', '_')
            justify_str         = container.style.justify_content.lower().replace('-', '_')
            flex_direction_val  = FlexDirection.ROW
            align_items_val     = AlignItems.START
            justify_content_val = JustifyContent.START

            width_pct  = parse_css_value(width_val)
            height_pct = parse_css_value(height_val)

            padding_val = parse_padding_values(container)
            margin_val  = parse_margin_values(container)
            border_val  = parse_border_values(container)

            if pos_str in ['absolute', 'fixed']:
                position_val = Position.ABSOLUTE
            elif pos_str in ['relative', 'static']:
                position_val = Position.RELATIVE

            if disp_str == 'none':
                display_val = Display.NONE
            elif disp_str == 'flex':
                display_val = Display.FLEX
            elif disp_str == 'grid':
                display_val = Display.GRID
            elif disp_str == 'block':
                display_val = Display.BLOCK

            if flex_dir_str == 'row':
                flex_direction_val = FlexDirection.ROW
            elif flex_dir_str == 'column':
                flex_direction_val = FlexDirection.COLUMN
            elif flex_dir_str == 'row_reverse':
                flex_direction_val = FlexDirection.ROW_REVERSE
            elif flex_dir_str == 'column_reverse':
                flex_direction_val = FlexDirection.COLUMN_REVERSE

            if align_str == 'start':
                align_items_val = AlignItems.START
            elif align_str == 'end':
                align_items_val = AlignItems.END
            elif align_str == 'flex_start':
                align_items_val = AlignItems.FLEX_START
            elif align_str == 'flex_end':
                align_items_val = AlignItems.FLEX_END
            elif align_str == 'center':
                align_items_val = AlignItems.CENTER
            elif align_str == 'baseline':
                align_items_val = AlignItems.BASELINE
            elif align_str == 'stretch':
                align_items_val = AlignItems.STRETCH

            if justify_str == 'start':
                justify_content_val = JustifyContent.START
            elif justify_str == 'end':
                justify_content_val = JustifyContent.END
            elif justify_str == 'flex_start':
                justify_content_val = JustifyContent.FLEX_START
            elif justify_str == 'flex_end':
                justify_content_val = JustifyContent.FLEX_END
            elif justify_str == 'center':
                justify_content_val = JustifyContent.CENTER
            elif justify_str == 'stretch':
                justify_content_val = JustifyContent.STRETCH
            elif justify_str == 'space_between':
                justify_content_val = JustifyContent.SPACE_BETWEEN
            elif justify_str == 'space_evenly':
                justify_content_val = JustifyContent.SPACE_EVENLY
            elif justify_str == 'space_around':
                justify_content_val = JustifyContent.SPACE_AROUND

            node = Node(
                display         = display_val,
                position        = position_val,
                box_sizing      = BoxSizing.BORDER,
                flex_direction  = flex_direction_val,
                align_items     = align_items_val,
                justify_content = justify_content_val,
                key             = container.id,
                size            = (width_pct, height_pct),
                padding         = padding_val,
                margin          = margin_val,
                border          = border_val,
            )
            
            for child in container.children:
                child_node = create_node(child)
                node.add(child_node)

            return node

        self.root_node = create_node(self.theme.root)
        self.root_node.compute_layout(canvas_size)
        self.canvas_size = canvas_size
        get_all_nodes(self.theme.root, self.root_node)

    def recompute_layout(self, canvas_size):
        """Recompute layout with new canvas size and regenerate flattened data"""
        global node_flat, node_flat_abs
        
        node_flat.clear()
        node_flat_abs.clear()
        
        self.root_node.compute_layout(canvas_size)
        self.canvas_size = canvas_size
        
        def get_all_nodes(container, node):
            border_box     = node.get_box(Edge.BORDER, relative=True)
            border_box_abs = node.get_box(Edge.BORDER, relative=False)
            content_box    = node.get_box(Edge.CONTENT, relative=True)
            content_box_abs = node.get_box(Edge.CONTENT, relative=False)
            padding_box    = node.get_box(Edge.PADDING, relative=True)
            margin_box     = node.get_box(Edge.MARGIN, relative=True)
            margin_box_abs = node.get_box(Edge.MARGIN, relative=False)
            
            edge_used, edge_used_abs = border_box, border_box_abs

            node_flat[container.id] = {
                'x'      : edge_used.x,
                'y'      : edge_used.y,
                'width'  : edge_used.width,
                'height' : edge_used.height
            }

            node_flat_abs[container.id] = {
                'x'      : edge_used_abs.x,
                'y'      : edge_used_abs.y,
                'width'  : edge_used_abs.width,
                'height' : edge_used_abs.height
            }
            
            for i, _container in enumerate(container.children):
                get_all_nodes(_container, node[i])
        
        get_all_nodes(self.theme.root, self.root_node)
        
        self.json_data = []
        self.abs_json_data = []
        self.flatten_node_tree()
        
        return self.abs_json_data

    def flatten_node_tree(self):
        container_id_to_index = {}

        container_id_to_index[self.theme.root.id] = 0

        def get_indices(container):
            for container in container.children:
                container_id_to_index[container.id] = len(container_id_to_index)
                get_indices(container)

        get_indices(self.theme.root)

        def flatten_conts(container):
            x = y = width = height = 0
            node         = node_flat.get(container.id)
            parent_index = -1
            if container.parent and container.parent.id in container_id_to_index:
                parent_index = container_id_to_index[container.parent.id]
            if node:
                x, y          = node['x'], node['y']
                width, height = node['width'], node['height']
                overflow_value = False if container.style.overflow == 'HIDDEN' else True
                container_data = {
                    "id"                       : container.id,
                    "style"                    : container.style.id,
                    "display"                  : True if container.style.display != Display.NONE else False,
                    "overflow"                 : overflow_value,
                    "data"                     : container.data,
                    "img"                      : container.img,
                    "aspect_ratio"             : container.style.aspect_ratio,
                    "text"                     : container.text,
                    "font"                     : container.font,
                    "position"                 : [x, y],
                    "size"                     : [width, height],
                    "color"                    : list(container.style.color),
                    "color_1"                  : list(container.style.color_1),
                    "color_gradient_rot"       : container.style.color_gradient_rot,
                    "hover_color"              : list(container.style.hover_color),
                    "hover_color_1"            : list(container.style.hover_color_1),
                    "hover_color_gradient_rot" : container.style.hover_color_gradient_rot,
                    "click_color"              : list(container.style.click_color),
                    "click_color_1"            : list(container.style.click_color_1),
                    "click_color_gradient_rot" : container.style.click_color_gradient_rot,
                    "border_color"             : list(container.style.border_color),
                    "border_color_1"           : list(container.style.border_color_1),
                    "border_color_gradient_rot": container.style.border_color_gradient_rot,
                    "border_radius"            : container.style.border_radius,
                    "border_width"             : container.style.border_width,
                    "text_color"               : list(container.style.text_color),
                    "text_color_1"             : list(container.style.text_color_1),
                    "text_color_gradient_rot"  : container.style.text_color_gradient_rot,
                    "text_scale"               : container.style.text_scale,
                    "text_x"                   : container.style.text_x,
                    "text_y"                   : container.style.text_y,
                    "box_shadow_color"         : list(container.style.box_shadow_color),
                    "box_shadow_offset"        : list(container.style.box_shadow_offset),
                    "box_shadow_blur"          : container.style.box_shadow_blur,
                    "parent"                   : parent_index,
                    "passive"                  : container.passive,
                    "children"                 : [container_id_to_index[child.id] for child in container.children] if container.children else [],
                    "click"                    : container.click,
                    "toggle"                   : container.toggle,
                    "scroll"                   : container.scroll,
                    "_scroll_value"            : container._scroll_value,
                    "hover"                    : container.hover,
                    "hoverout"                 : container.hoverout,
                    "_clicked"                 : False,
                    "_prev_clicked"            : False,
                    "_toggle_value"            : False,
                    "_toggled"                 : False,
                    "_prev_toggled"            : False,
                    "_hovered"                 : False,
                    "_prev_hovered"            : False
                }
                self.json_data.append(container_data)
            for child in container.children:
                flatten_conts(child)

        flatten_conts(self.theme.root)

        def flatten_conts_abs(container):
            x = y = width = height = 0
            node         = node_flat_abs.get(container.id)
            parent_index = -1
            if container.parent and container.parent.id in container_id_to_index:
                parent_index = container_id_to_index[container.parent.id]
            if node:
                x, y          = node['x'], node['y']
                width, height = node['width'], node['height']
                container_data = {
                    "id"                       : container.id,
                    "style"                    : container.style.id,
                    "display"                  : True if container.style.display != Display.NONE else False,
                    "overflow"                 : False if container.style.overflow == 'HIDDEN' else True,
                    "data"                     : container.data,
                    "img"                      : container.img,
                    "aspect_ratio"             : container.style.aspect_ratio,
                    "text"                     : container.text,
                    "font"                     : container.font,
                    "position"                 : [x, y],
                    "size"                     : [width, height],
                    "color"                    : list(container.style.color),
                    "color_1"                  : list(container.style.color_1),
                    "color_gradient_rot"       : container.style.color_gradient_rot,
                    "hover_color"              : list(container.style.hover_color),
                    "hover_color_1"            : list(container.style.hover_color_1),
                    "hover_color_gradient_rot" : container.style.hover_color_gradient_rot,
                    "click_color"              : list(container.style.click_color),
                    "click_color_1"            : list(container.style.click_color_1),
                    "click_color_gradient_rot" : container.style.click_color_gradient_rot,
                    "border_color"             : list(container.style.border_color),
                    "border_color_1"           : list(container.style.border_color_1),
                    "border_color_gradient_rot": container.style.border_color_gradient_rot,
                    "border_radius"            : container.style.border_radius,
                    "border_width"             : container.style.border_width,
                    "text_color"               : list(container.style.text_color),
                    "text_color_1"             : list(container.style.text_color_1),
                    "text_color_gradient_rot"  : container.style.text_color_gradient_rot,
                    "text_scale"               : container.style.text_scale,
                    "text_x"                   : container.style.text_x,
                    "text_y"                   : container.style.text_y,
                    "box_shadow_color"         : list(container.style.box_shadow_color),
                    "box_shadow_offset"        : list(container.style.box_shadow_offset),
                    "box_shadow_blur"          : container.style.box_shadow_blur,
                    "parent"                   : parent_index,
                    "passive"                  : container.passive,
                    "children"                 : [container_id_to_index[child.id] for child in container.children] if container.children else [],
                    "click"                    : container.click,
                    "toggle"                   : container.toggle,
                    "scroll"                   : container.scroll,
                    "_scroll_value"            : container._scroll_value,
                    "hover"                    : container.hover,
                    "hoverout"                 : container.hoverout,
                    "_clicked"                 : False,
                    "_prev_clicked"            : False,
                    "_toggle_value"            : False,
                    "_toggled"                 : False,
                    "_prev_toggled"            : False,
                    "_hovered"                 : False,
                    "_prev_hovered"            : False
                }
                self.abs_json_data.append(container_data)
            for child in container.children:
                flatten_conts_abs(child)

        flatten_conts_abs(self.theme.root)
 


import importlib
import importlib.util
import os
from time import sleep

class Compiler():
    def __init__(self, ui):
        self.ui = ui
    def compile(self):
        global global_vars
        from . import get_addon_root
        addon_dir = get_addon_root()
        
        for _script_ in self.ui.theme.scripts:
            module_name = _script_.replace(".py", "")
            try:
                script_path = os.path.join(addon_dir, f"{module_name}.py")
                spec = importlib.util.spec_from_file_location(module_name, script_path)
                if spec and spec.loader:
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                    if hasattr(module, 'main'):
                        self.ui = module.main(self, self.ui)
                sleep(0.1)
            except (ImportError, FileNotFoundError) as e:
                print(f"Failed to import {module_name}: {e}")
        return self.ui



import bpy
import os
from .parser    import UI
from .compiler  import Compiler
from .extract_images import ImageExtractor
from .extract_text   import TextExtractor
from .extract_text_input import TextInputExtractor

XWZ_UI                = None
text_blocks           = {}
text_input_blocks     = {}
image_blocks          = {}
image_blocks_relative = {}
_container_json_data  = []

class XWZ_OT_ui_parser(bpy.types.Operator): 
    bl_idname = "xwz.parse_app_ui"
    bl_label  = "Parse App UI"

    conf_path : bpy.props.StringProperty(
        name        = "UI Config File Path",
        description = "Path to the configuration file for the UI layout",
        default     = "//index.toml"
    )

    def print_ui_struct(self):
        def find_children(children):
            for child in children:
                #print all child attributes
                print(f"Container ID:[ {child.id} ]")
                print(' ├────────────────────────────────────────────')
                print(f" ├─ Data               : {child.data}")
                print(f" ├─ Display            : {child.display}")
                print(f" ├─ Image              : {child.img}")
                print(f" ├─ Aspect Ratio       : {child.aspect_ratio}")
                print(f" ├─ Overflow           : {child.overflow}")
                print(f" ├─ Style              : {child.style}")
                print(' ├────────────────────────────────────────────')
                print(f" ├─ Parent ID          : [ {child.parent.id if child.parent else 'None'} ]")
                print(f" ├─ Number of Children : {len(child.children)}")
                if len(child.children) > 0:
                    for cc in child.children:
                        print(f" ├─────── Child ID : [ {cc.id} ]")
                print(' ├────────────────────────────────────────────')
                print(f" ├─  Position           : ({child.x}, {child.y})")
                print(f" ├─  Size               : ({child.width}, {child.height})")
                print(f" ├─  Color              : {child.color}")
                print(f" ├─  Color1             : {child.color_1}")
                print(f" ├─  Color G Rotation   : {child.color_gradient_rot}")
                print(f" ├─  Hover Color        : {child.hover_color}")
                print(f" ├─  Hover Color1       : {child.hover_color_1}")
                print(f" ├─  Hover Color G Rot  : {child.hover_color_gradient_rot}")
                print(f" ├─  Click Color        : {child.click_color}")
                print(f" ├─  Click Color1       : {child.click_color_1}")
                print(f" ├─  Click Color G Rot  : {child.click_color_gradient_rot}")
                print(f" ├─  Border Radius      : {child.border_radius}")
                print(f" ├─  Border Width       : {child.border_width}")
                print(f" ├─  Border Color       : {child.border_color}")
                print(f" ├─  Border Color1      : {child.border_color_1}")
                print(f" ├─  Border Color G Rot : {child.border_color_gradient_rot}")
                print(' ├───────────────────────────────────────────')
                print(f" ├─  Click Events       : {child.click}")
                print(f" ├─  Toggle Events      : {child.toggle}")
                print(f" ├─  Scroll Events      : {child.scroll}")
                print(f" ├─  Hover Events       : {child.hover}")
                print(' ├───────────────────────────────────────────')
                print(f" ├─  Font             : {child.font}")
                print(f" ├─  Text             : {child.text}")
                print(f" ├─  Text Scale       : {child.text_scale}")
                print(f" ├─  Text X           : {child.text_x}")
                print(f" ├─  Text Y           : {child.text_y}")
                print(f" ├─  Text Color       : {child.text_color}")
                print(f" ├─  Text Color1      : {child.text_color_1}")
                print(f" ├─  Text Color G Rot : {child.text_color_gradient_rot}")
                print(' ├───────────────────────────────────────────')
                print(f" ├─  Box Shadow Color   : {child.box_shadow_color}")
                print(f" ├─  Box Shadow Offset  : {child.box_shadow_offset}")
                print(f" ├─  Box Shadow Blur    : {child.box_shadow_blur}")
                print(' └───────────────────────────────────────────')
                find_children(child.children)
        find_children(self.ui.theme.root.children)
    
    def dump_ui_struct(self):
        global _container_json_data
        self.container_json_data = self.ui.abs_json_data
        _container_json_data = self.container_json_data
        return
    
    def execute(self, context):
        # get viewport size
        region_size = (800, 600)
        for area in context.screen.areas:
            if area.type == 'VIEW_3D':
                for region in area.regions:
                    if region.type == 'WINDOW':
                        region_size = (region.width, region.height)
                        break
                break
        global XWZ_UI, text_blocks, text_input_blocks, image_blocks, image_blocks_relative
        from . import get_addon_root
        addon_dir  = get_addon_root()

        self.ui              = UI(os.path.join(addon_dir, self.conf_path), addon_dir, canvas_size=region_size)
        self.compiler        = Compiler(self.ui)
        self.ui              = self.compiler.compile()
        
        self.image_extractor = ImageExtractor(self.ui, self.ui.abs_json_data)
        self.text_extractor  = TextExtractor(self.ui, self.ui.abs_json_data)
        self.text_input_extractor = TextInputExtractor(self.ui, self.ui.abs_json_data)

        text_blocks           = self.text_extractor.text_blocks
        text_input_blocks     = self.text_input_extractor.text_input_blocks
        image_blocks          = self.image_extractor.image_blocks
        image_blocks_relative = self.image_extractor.image_blocks_relative

        XWZ_UI = self.ui  # Store UI instance globally for layout recomputation
        self.dump_ui_struct()
        return {'FINISHED'}

def recompute_layout(canvas_size):
    global XWZ_UI, _container_json_data, text_blocks, text_input_blocks, image_blocks, image_blocks_relative
    
    if XWZ_UI is None:
        return None
    
    # Recompute layout with new canvas size
    updated_data = XWZ_UI.recompute_layout(canvas_size)
    
    # Update global container data
    _container_json_data = updated_data
    
    # Update text and image blocks with new layout
    text_extractor = TextExtractor(XWZ_UI, XWZ_UI.abs_json_data)
    text_input_extractor = TextInputExtractor(XWZ_UI, XWZ_UI.abs_json_data)
    image_extractor = ImageExtractor(XWZ_UI, XWZ_UI.abs_json_data)
    
    text_blocks = text_extractor.text_blocks
    text_input_blocks = text_input_extractor.text_input_blocks
    image_blocks = image_extractor.image_blocks
    image_blocks_relative = image_extractor.image_blocks_relative
    
    return _container_json_data

def sync_dirty_containers():
    """
    Sync container state from live Container objects to _container_json_data
    without full layout recomputation. This is called when scripts modify
    container properties (like colors, text, images) via event handlers.
    """
    global XWZ_UI, _container_json_data, text_blocks, text_input_blocks, image_blocks, image_blocks_relative
    
    if XWZ_UI is None or not _container_json_data:
        return False
    
    # Check if any containers are dirty
    has_dirty = check_dirty_containers(XWZ_UI.theme.root)
    if not has_dirty:
        return False
    
    # Sync live Container tree state to flattened json data
    # by reusing the existing abs_json_data structure
    XWZ_UI.abs_json_data = []
    XWZ_UI.flatten_node_tree()
    _container_json_data = XWZ_UI.abs_json_data
    
    # Regenerate text and image blocks to pick up changes
    # This ensures that changes to container.text and container.img
    # are reflected in the rendered output
    text_extractor = TextExtractor(XWZ_UI, XWZ_UI.abs_json_data)
    text_input_extractor = TextInputExtractor(XWZ_UI, XWZ_UI.abs_json_data)
    image_extractor = ImageExtractor(XWZ_UI, XWZ_UI.abs_json_data)
    
    text_blocks = text_extractor.text_blocks
    text_input_blocks = text_input_extractor.text_input_blocks
    image_blocks = image_extractor.image_blocks
    image_blocks_relative = image_extractor.image_blocks_relative
    
    # Clear dirty flags
    clear_dirty_flags(XWZ_UI.theme.root)
    
    return True

def check_dirty_containers(container):
    """Recursively check if any container in tree is marked dirty"""
    if hasattr(container, '_dirty') and container._dirty:
        return True
    for child in container.children:
        if check_dirty_containers(child):
            return True
    return False

def clear_dirty_flags(container):
    """Recursively clear all dirty flags in container tree"""
    if hasattr(container, '_dirty'):
        container._dirty = False
    for child in container.children:
        clear_dirty_flags(child)
    


import os
import bpy
import gpu
from gpu_extras.batch import batch_for_shader
from mathutils import Matrix

_image_instances = []
_draw_handle = None

class ImageManager:
    _instance = None
    
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance
    
    def __init__(self):
        if not self._initialized:
            self.images = {}
            self.textures = {}
            self._load_images()
            self._initialized = True
    
    def _load_images(self):
        from . import get_addon_root
        addon_assets_path = os.path.join(get_addon_root(), "assets")
        if os.path.exists(addon_assets_path):
            for image_file in os.listdir(addon_assets_path):
                if image_file.lower().endswith(('.png', '.jpg', '.jpeg', '.bmp', '.tiff', '.tga', '.webp')):
                    image_path = os.path.join(addon_assets_path, image_file)
                    try:
                        if image_file not in bpy.data.images:
                            bpy_image = bpy.data.images.load(image_path)
                        else:
                            bpy_image = bpy.data.images[image_file]
                        
                        bpy_image.alpha_mode = 'PREMUL'
                        
                        texture = gpu.texture.from_image(bpy_image)
                        image_name = os.path.splitext(image_file)[0]
                        self.images[image_name] = image_path
                        self.textures[image_name] = texture
                    except Exception as e:
                        print(f"Failed to load image {image_file}: {e}")
        
    def get_texture(self, image_name):
        return self.textures.get(image_name, None)
    
    def get_available_images(self):
        return list(self.images.keys())
    
    def unload_images(self):
        for image_name, image_path in self.images.items():
            try:
                image_file = os.path.basename(image_path)
                if image_file in bpy.data.images:
                    bpy.data.images.remove(bpy.data.images[image_file])
            except Exception as e:
                print(f"Failed to remove image {image_name}: {e}")
        
        self.textures.clear()
        self.images.clear()
    
    def reload_images(self):
        """Reload all images - used when addon is re-enabled without Blender restart"""
        self.unload_images()
        self._load_images()
    
    @classmethod
    def reset_instance(cls):
        """Reset the singleton instance - used during addon unregister"""
        if cls._instance is not None:
            if cls._instance._initialized:
                cls._instance.unload_images()
            cls._instance = None

image_manager = ImageManager()

class ImageInstance:
    def __init__(self, container_id, image_name=None, pos=[50, 50], size=[100, 100], mask=None, aspect_ratio=True, align_h='LEFT', align_v='TOP'):
        self.id           = len(_image_instances)
        self.container_id = container_id
        self.image_name   = image_name
        self.texture      = image_manager.get_texture(self.image_name) if self.image_name else None
        self.position     = pos
        self.size         = size
        self.mask         = mask
        self.aspect_ratio = aspect_ratio
        self.align_h      = align_h
        self.align_v      = align_v
        self.shader       = gpu.shader.from_builtin('IMAGE')
        self.batch        = None
        self._create_batch()
    
    def _create_batch(self):
        if self.texture:
            vertices = [
                (0, 0), (1, 0), (1, 1), (0, 1)
            ]
            uvs = [
                (0, 0), (1, 0), (1, 1), (0, 1)
            ]
            indices = [(0, 1, 2), (0, 2, 3)]
            self.batch = batch_for_shader(
                self.shader, 'TRIS',
                {"pos": vertices, "texCoord": uvs},
                indices=indices
            )
    
    def get_display_size(self):
        if not self.aspect_ratio or not self.texture:
            return self.size
        
        tex_width = self.texture.width
        tex_height = self.texture.height
        
        if tex_width == 0 or tex_height == 0:
            return self.size
        
        tex_aspect = tex_width / tex_height
        target_width, target_height = self.size
        target_aspect = target_width / target_height
        
        if tex_aspect > target_aspect:
            actual_width = target_width
            actual_height = target_width / tex_aspect
        else:
            actual_width = target_height * tex_aspect
            actual_height = target_height
        
        return [actual_width, actual_height]

    def update_image(self, new_image_name):
        if new_image_name in image_manager.get_available_images():
            self.image_name = new_image_name
            self.texture = image_manager.get_texture(new_image_name)
            self._create_batch()
            self._trigger_redraw()
    
    def update_size(self, new_size):
        self.size = [max(1, min(2000, new_size[0])), max(1, min(2000, new_size[1]))]
        self._trigger_redraw()
    
    def update_position(self, new_pos):
        self.position = list(new_pos)
        self._trigger_redraw()
    
    def update_mask(self, new_mask):
        self.mask = new_mask
        self._trigger_redraw()
    
    def update_aspect_ratio(self, new_aspect_ratio):
        self.aspect_ratio = new_aspect_ratio
        self._trigger_redraw()
    
    def update_all(self, image_name=None, size=None, pos=None, mask=None, aspect_ratio=None, align_h=None, align_v=None):
        if image_name is not None and image_name in image_manager.get_available_images():
            self.image_name = image_name
            self.texture = image_manager.get_texture(image_name)
            self._create_batch()
        if size is not None:
            self.size = [max(1, min(2000, size[0])), max(1, min(2000, size[1]))]
        if pos is not None:
            self.position = list(pos)
        if mask is not None:
            self.mask = mask
        if aspect_ratio is not None:
            self.aspect_ratio = aspect_ratio
        if align_h is not None:
            self.align_h = align_h
        if align_v is not None:
            self.align_v = align_v
        self._trigger_redraw()
    
    def _trigger_redraw(self):
        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                area.tag_redraw()

def draw_all_images():
    viewport_height = 0
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for region in area.regions:
                if region.type == 'WINDOW':
                    viewport_height = region.height
                    break
            break
    
    gpu.state.blend_set('ALPHA_PREMULT')
    
    for instance in _image_instances:
        if not instance.texture or not instance.batch:
            continue
        
        if instance.mask and instance.mask[2] > 0 and instance.mask[3] > 0:
            xmin = instance.mask[0]
            ymin = viewport_height - instance.mask[1] - instance.mask[3]
            xmax = instance.mask[0] + instance.mask[2]
            ymax = viewport_height - instance.mask[1]
            gpu.state.scissor_test_set(True)
            gpu.state.scissor_set(int(xmin), int(ymin), int(xmax - xmin), int(ymax - ymin))
        
        display_size = instance.get_display_size()
        
        x_pos = instance.position[0]
        y_pos = instance.position[1]
        
        if instance.mask and instance.mask[2] > 0 and instance.mask[3] > 0:
            container_width = instance.mask[2]
            container_height = instance.mask[3]
            
            if instance.align_h == 'LEFT':
                x_pos = instance.mask[0]
            elif instance.align_h == 'CENTER':
                x_pos = instance.mask[0] + (container_width - display_size[0]) / 2
            elif instance.align_h == 'RIGHT':
                x_pos = instance.mask[0] + container_width - display_size[0]
            
            if instance.align_v == 'TOP':
                y_pos = instance.mask[1]
            elif instance.align_v == 'CENTER':
                y_pos = instance.mask[1] + (container_height - display_size[1]) / 2
            elif instance.align_v == 'BOTTOM':
                y_pos = instance.mask[1] + container_height - display_size[1]
        
        flipped_y = viewport_height - y_pos - display_size[1]
        
        scale_matrix = Matrix.Diagonal((display_size[0], display_size[1], 1.0, 1.0))
        translation_matrix = Matrix.Translation((x_pos, flipped_y, 0))
        
        matrix = gpu.matrix.get_projection_matrix()
        matrix = matrix @ translation_matrix @ scale_matrix
        
        instance.shader.bind()
        instance.shader.uniform_sampler("image", instance.texture)
        
        gpu.matrix.push_projection()
        gpu.matrix.load_projection_matrix(matrix)
        
        instance.batch.draw(instance.shader)
        
        gpu.matrix.pop_projection()
        
        if instance.mask and instance.mask[2] > 0 and instance.mask[3] > 0:
            gpu.state.scissor_test_set(False)
    
    gpu.state.blend_set('NONE')

class DrawImageOP(bpy.types.Operator):
    bl_idname = "xwz.draw_image"
    bl_label = "Add Image Instance"
    
    def get_image_items(self, context):
        image_manager._load_images()
        items = [(name, name, "") for name in image_manager.get_available_images()]
        return items if items else [("none", "None", "")]
    
    container_id: bpy.props.StringProperty(name="Container ID", default="root")
    image_name  : bpy.props.EnumProperty(
        name  = "Image",
        items = get_image_items
    )
    width       : bpy.props.IntProperty(name="Width", default=100, min=1, max=2000)
    height      : bpy.props.IntProperty(name="Height", default=100, min=1, max=2000)
    x_pos       : bpy.props.IntProperty(name="X Position", default=50)
    y_pos       : bpy.props.IntProperty(name="Y Position", default=50)
    mask_x      : bpy.props.IntProperty(name="Mask X", default=0)
    mask_y      : bpy.props.IntProperty(name="Mask Y", default=0)
    mask_width  : bpy.props.IntProperty(name="Mask Width", default=0)
    mask_height : bpy.props.IntProperty(name="Mask Height", default=0)
    aspect_ratio: bpy.props.BoolProperty(name="Keep Aspect Ratio", default=True)
    align_h     : bpy.props.EnumProperty(
        name="Horizontal Align",
        items=[('LEFT', 'Left', ''), ('CENTER', 'Center', ''), ('RIGHT', 'Right', '')],
        default='LEFT'
    )
    align_v     : bpy.props.EnumProperty(
        name="Vertical Align",
        items=[('TOP', 'Top', ''), ('CENTER', 'Center', ''), ('BOTTOM', 'Bottom', '')],
        default='TOP'
    )
    
    def execute(self, context):
        global _draw_handle, _image_instances
        
        mask = None
        if self.mask_width > 0 and self.mask_height > 0:
            mask = [self.mask_x, self.mask_y, self.mask_width, self.mask_height]
        
        new_instance = ImageInstance(
            container_id=self.container_id,
            image_name=self.image_name,
            pos=[self.x_pos, self.y_pos],
            size=[self.width, self.height],
            mask=mask,
            aspect_ratio=self.aspect_ratio,
            align_h=self.align_h,
            align_v=self.align_v
        )
        _image_instances.append(new_instance)
        
        if _draw_handle is None:
            _draw_handle = bpy.types.SpaceView3D.draw_handler_add(
                draw_all_images, (), 'WINDOW', 'POST_PIXEL')
        
        context.area.tag_redraw()
        self.report({'INFO'}, f"Added image instance #{new_instance.id} with image {self.image_name}")
        return {'FINISHED'}

class RemoveImageOP(bpy.types.Operator):
    bl_idname = "xwz.remove_image"
    bl_label = "Remove Image Instance"
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    
    def execute(self, context):
        global _draw_handle, _image_instances
        
        for i, instance in enumerate(_image_instances):
            if instance.id == self.instance_id:
                _image_instances.pop(i)
                self.report({'INFO'}, f"Removed image instance #{self.instance_id}")
                break
        else:
            self.report({'ERROR'}, f"Image instance #{self.instance_id} not found")
            return {'CANCELLED'}
        
        if not _image_instances and _draw_handle is not None:
            bpy.types.SpaceView3D.draw_handler_remove(_draw_handle, 'WINDOW')
            _draw_handle = None
        
        context.area.tag_redraw()
        return {'FINISHED'}

class ClearImageOP(bpy.types.Operator):
    bl_idname = "xwz.clear_images"
    bl_label = "Clear All Images"
    
    def execute(self, context):
        global _draw_handle, _image_instances
        
        _image_instances.clear()
        
        if _draw_handle is not None:
            bpy.types.SpaceView3D.draw_handler_remove(_draw_handle, 'WINDOW')
            _draw_handle = None
        
        context.area.tag_redraw()
        return {'FINISHED'}

class UpdateImageOP(bpy.types.Operator):
    bl_idname = "xwz.update_image"
    bl_label = "Update Image Instance"
    
    def get_image_items(self, context):
        image_manager._load_images()
        # Add a "no change" option at the beginning
        items = [("__NOCHANGE__", "No Change", "Don't change the image")]
        items.extend([(name, name, "") for name in image_manager.get_available_images()])
        return items
    
    instance_id: bpy.props.IntProperty(name="Instance ID", default=0, min=0)
    image_name : bpy.props.EnumProperty(
        name="Image",
        items=get_image_items,
        default=0  # Will default to "No Change"
    )
    width      : bpy.props.IntProperty(name="Width", default=-1, min=-1, max=2000)  # -1 = no change
    height     : bpy.props.IntProperty(name="Height", default=-1, min=-1, max=2000)  # -1 = no change
    x_pos      : bpy.props.IntProperty(name="X Position", default=-999999)  # sentinel = no change
    y_pos      : bpy.props.IntProperty(name="Y Position", default=-999999)  # sentinel = no change
    mask_x     : bpy.props.IntProperty(name="Mask X", default=-999999)  # sentinel = no change
    mask_y     : bpy.props.IntProperty(name="Mask Y", default=-999999)  # sentinel = no change
    mask_width : bpy.props.IntProperty(name="Mask Width", default=-1, min=-1)  # -1 = no change
    mask_height: bpy.props.IntProperty(name="Mask Height", default=-1, min=-1)  # -1 = no change
    aspect_ratio: bpy.props.EnumProperty(
        name="Keep Aspect Ratio",
        items=[
            ("__NOCHANGE__", "No Change", "Don't change aspect ratio setting"),
            ("TRUE", "True", "Keep aspect ratio"),
            ("FALSE", "False", "Don't keep aspect ratio")
        ],
        default=0  # Will default to "No Change"
    )
    align_h: bpy.props.EnumProperty(
        name="Horizontal Align",
        items=[('__NOCHANGE__', 'No Change', ''), ('LEFT', 'Left', ''), ('CENTER', 'Center', ''), ('RIGHT', 'Right', '')],
        default='__NOCHANGE__'
    )
    align_v: bpy.props.EnumProperty(
        name="Vertical Align",
        items=[('__NOCHANGE__', 'No Change', ''), ('TOP', 'Top', ''), ('CENTER', 'Center', ''), ('BOTTOM', 'Bottom', '')],
        default='__NOCHANGE__'
    )
    
    def execute(self, context):
        for instance in _image_instances:
            if instance.id == self.instance_id:
                kwargs = {}
                
                # Check image name
                if self.image_name != "__NOCHANGE__" and self.image_name in image_manager.get_available_images():
                    kwargs['image_name'] = self.image_name
                
                # Check size
                if self.width != -1 or self.height != -1:
                    # If only one dimension is specified, keep the other unchanged
                    new_width = max(1, min(2000, self.width)) if self.width != -1 else instance.size[0]
                    new_height = max(1, min(2000, self.height)) if self.height != -1 else instance.size[1]
                    kwargs['size'] = [new_width, new_height]
                
                # Check position
                if self.x_pos != -999999 or self.y_pos != -999999:
                    # If only one coordinate is specified, keep the other unchanged
                    new_x = self.x_pos if self.x_pos != -999999 else instance.position[0]
                    new_y = self.y_pos if self.y_pos != -999999 else instance.position[1]
                    kwargs['pos'] = [new_x, new_y]
                
                # Check mask - only update if any mask property is specified
                if (self.mask_x != -999999 or self.mask_y != -999999 or 
                    self.mask_width != -1 or self.mask_height != -1):
                    
                    current_mask = instance.mask or [0, 0, 0, 0]
                    new_mask_x = self.mask_x if self.mask_x != -999999 else current_mask[0]
                    new_mask_y = self.mask_y if self.mask_y != -999999 else current_mask[1]
                    new_mask_w = self.mask_width if self.mask_width != -1 else current_mask[2]
                    new_mask_h = self.mask_height if self.mask_height != -1 else current_mask[3]
                    
                    if new_mask_w > 0 and new_mask_h > 0:
                        kwargs['mask'] = [new_mask_x, new_mask_y, new_mask_w, new_mask_h]
                    else:
                        kwargs['mask'] = None
                
                # Check aspect ratio
                if self.aspect_ratio != "__NOCHANGE__":
                    kwargs['aspect_ratio'] = self.aspect_ratio == "TRUE"
                
                if self.align_h != '__NOCHANGE__':
                    kwargs['align_h'] = self.align_h
                
                if self.align_v != '__NOCHANGE__':
                    kwargs['align_v'] = self.align_v
                
                if kwargs:
                    instance.update_all(**kwargs)
                    updated_props = list(kwargs.keys())

                else:
                    self.report({'INFO'}, f"No properties specified to update for image instance #{self.instance_id}")
                
                return {'FINISHED'}
        
        self.report({'ERROR'}, f"Image instance #{self.instance_id} not found")
        return {'CANCELLED'}

def register():
    global image_manager
    
    # Ensure images are loaded/reloaded when addon is (re)enabled
    if image_manager is None:
        # After unregister was called - recreate the singleton
        image_manager = ImageManager()
    elif image_manager._initialized:
        # Addon was previously loaded - reload images
        image_manager.reload_images()
    
    bpy.utils.register_class(DrawImageOP)
    bpy.utils.register_class(RemoveImageOP)
    bpy.utils.register_class(ClearImageOP)
    bpy.utils.register_class(UpdateImageOP)

def unregister():
    global _draw_handle, _image_instances, image_manager
    
    # Force clear all image instances
    _image_instances.clear()
    
    if _draw_handle is not None:
        bpy.types.SpaceView3D.draw_handler_remove(_draw_handle, 'WINDOW')
        _draw_handle = None
    
    # Unload images and reset the singleton
    ImageManager.reset_instance()
    image_manager = None
    
    bpy.utils.unregister_class(DrawImageOP)
    bpy.utils.unregister_class(RemoveImageOP)
    bpy.utils.unregister_class(ClearImageOP)
    bpy.utils.unregister_class(UpdateImageOP)


import bpy
from bpy.types import Panel
from . import render

class XWZ_PT_panel(Panel):
    bl_label       = "puree"
    bl_idname      = "XWZ_PT_panel"
    bl_space_type  = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category    = "puree"
    
    @classmethod
    def poll(cls, context):
        return context.window_manager.xwz_debug_panel
    
    def draw(self, context):
        layout = self.layout
        
        if render._render_data and render._render_data.running:
            layout.label(text="Running", icon='PLAY')
            
            box = layout.box()
            col = box.column(align=True)
            col.separator()
            col.label(text=f"Texture: {render._render_data.texture_size[0]}x{render._render_data.texture_size[1]}")
            col.label(text=f"Compute FPS: {render._render_data.compute_fps:.1f}")
            
            layout.separator()
            layout.operator("xwz.stop_ui", icon='PAUSE')
        else:
            layout.label(text="Paused", icon='PAUSE')
            layout.operator("xwz.start_ui", icon='PLAY')
        
        layout.separator()
        layout.label(text="This is a debug panel.")

def register():
    bpy.utils.register_class(XWZ_PT_panel)

def unregister():
    bpy.utils.unregister_class(XWZ_PT_panel)


from __future__ import annotations
from typing import Optional, List

class Style(): 
    def __init__(self): 
        self.id: str = ""

        self.width  : float = 0.0
        self.height : float = 0.0

        self.color              : List[float] = [0.0, 0.0, 0.0, 1.0]
        self.color_1            : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.color_gradient_rot : float       = 0.0
        
        self.hover_color             : List[float] = [0.0, 0.0, 0.0, -1.0]
        self.hover_color_1           : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.hover_color_gradient_rot: float       = 0.0

        self.click_color             : List[float] = [0.0, 0.0, 0.0, -1.0]
        self.click_color_1           : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.click_color_gradient_rot: float       = 0.0
        
        self.text_x                  : float       = 0.0
        self.text_y                  : float       = 0.0
        self.text_scale              : float       = 12.0
        self.text_color              : List[float] = [1.0, 1.0, 1.0, 1.0]
        self.text_color_1            : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.text_color_gradient_rot: float        = 0.0
        self.text_align_h            : str         = 'LEFT'
        self.text_align_v            : str         = 'CENTER'

        self.img_align_h            : str         = 'LEFT'
        self.img_align_v            : str         = 'TOP'

        self.border_radius            : float       = 0.0
        self.border_width             : float       = 0.0
        self.border_color             : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.border_color_1           : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.border_color_gradient_rot: float       = 0.0
        
        self.box_shadow_color : List[float] = [0.0, 0.0, 0.0, 0.0]
        self.box_shadow_offset: List[float] = [0.0, 0.0, 0.0]
        self.box_shadow_blur  : float       = 0.0

        # STRECHABLE PROPERTIES -------------------------------------------------------------------------------------------
        self.display         : str  = 'FLEX'
        self.overflow        : str  = 'HIDDEN'
        self.scrollbar_width : float = 0.0
        self.position        : str  = 'RELATIVE'
        
        self.align_items    : str = 'START'
        self.justify_items  : str = 'START'
        self.align_self     : str = 'START'
        self.justify_self   : str = 'START'
        self.align_content  : str = 'START'
        self.justify_content: str = 'START'
        
        self.size         : List[float]     = [0.0, 0.0]
        self.min_size     : List[float]     = [0.0, 0.0]
        self.max_size     : List[float]     = [0.0, 0.0]
        self.aspect_ratio : bool = True
        
        self.flex_wrap     : str   = 'NO_WRAP'
        self.flex_direction: str   = 'ROW'
        self.flex_grow     : float = 0.0
        self.flex_shrink   : float = 1.0
        self.flex_basis    : float = 0.0
        
        self.grid_auto_flow       : str            = 'ROW'
        self.grid_template_rows   : Optional[List] = None
        self.grid_template_columns: Optional[List] = None
        self.grid_auto_rows       : Optional[List] = None
        self.grid_auto_columns    : Optional[List] = None
        self.grid_row             : str            = 'AUTO'
        self.grid_column          : str            = 'AUTO'

